import torch
from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from torch import sigmoid


class XorNeuralNet(torch.nn.Module):
    def __init__(self):
        super().__init__()
        # sigmoid(first layer) represents 'or' and 'and' of inputs respectively.
        self.layer1 = torch.nn.Linear(2, 2)
        self.layer1.weight.data = torch.tensor([[10.0, 10.0], [10.0, 10.0]])
        self.layer1.bias.data = torch.tensor([-5.0, -15.0])

        # sigmoid(second layer) represents values for xor = <or(inputs) and not(and(inputs))>
        self.layer2 = torch.nn.Linear(2, 1)
        self.layer2.weight.data = torch.tensor([[-10.0, 15.0], [10.0, -15.0]])
        self.layer2.bias.data = torch.tensor([5.0, -5.0])

    def forward(self, x):
        if not torch.is_floating_point(x):
            x = x.float()
        x = self.layer1(x)
        x = sigmoid(x)
        x = self.layer2(x)
        x = sigmoid(x)
        return x

    def randomize(self):
        # TODO
        pass

    def randomized_copy(self):
        return self

    def __repr__(self):
        return "xor neural net"


def get_data():
    p = IntegerVariable("p", 2)
    q = IntegerVariable("q", 2)
    xor = IntegerVariable("xor", 2)
    neural_factor = NeuralFactor(XorNeuralNet(), [p, q], xor)
    return neural_factor, p, q, xor


def test_potentials():
    neural_factor, p, q, xor = get_data()
    print(neural_factor.table_factor)
    assert neural_factor.table_factor == PyTorchTableFactor(
        [p, q, xor],
        [[[0.9928, 0.0072], [0.0079, 0.9921]], [[0.0079, 0.9921], [0.9999, 0.0001]]],
    )


def test_conditioning():
    neural_factor, p, q, xor = get_data()

    nothing = neural_factor.condition({})
    assert nothing == neural_factor

    not_p = neural_factor.condition({p: 0})
    print("xor | not p:", not_p)
    print("xor | not p table:", not_p.table_factor)
    assert not_p.table_factor == PyTorchTableFactor(
        [q, xor], [[0.9928, 0.0072], [0.0079, 0.9921]]
    )

    by_p = neural_factor.condition({p: 1})
    print("xor |     p:", by_p)
    print("xor |     p table:", by_p.table_factor)
    assert by_p.table_factor == PyTorchTableFactor(
        [q, xor], [[0.0079, 0.9921], [0.9999, 0.0001]]
    )

    p_and_q = neural_factor.condition({p: 1, q: 1})
    print("xor | p and q:", p_and_q)
    print("xor | p and q table:", p_and_q.table_factor)
    assert p_and_q.table_factor == PyTorchTableFactor([xor], [0.9999, 0.0001])

    xor_is_true = neural_factor.condition({xor: 1})
    print("xor | xor:", xor_is_true)
    print("xor | xor table:", xor_is_true.table_factor)
    assert xor_is_true.table_factor == PyTorchTableFactor(
        [p, q], [[0.0072, 0.9921], [0.9921, 0.0]]
    )

    xor_is_false = neural_factor.condition({xor: 0})
    print("xor | not xor:", xor_is_false)
    print("xor | not xor table:", xor_is_false.table_factor)
    assert xor_is_false.table_factor == PyTorchTableFactor(
        [p, q], [[0.9928, 0.008], [0.008, 0.999]]
    )

    p_and_xor_are_true = neural_factor.condition({p: 1, xor: 1})
    print("xor | p and xor:", p_and_xor_are_true)
    print("xor | p and xor table:", p_and_xor_are_true.table_factor)
    assert p_and_xor_are_true.table_factor == PyTorchTableFactor([q], [0.9921, 0.0])

    all_true = neural_factor.condition({p: 1, q: 1, xor: 1})
    print("xor | all true:", all_true)
    print("xor | all true table:", all_true.table_factor)
    assert all_true.table_factor == PyTorchTableFactor([], 0.0)


def test_normalize():
    neural_factor, p, q, xor = get_data()
    print("neural_factor.normalize()", neural_factor.normalize())
    assert neural_factor.normalize() == PyTorchTableFactor(
        [p, q, xor],
        [[[0.2482, 0.0018], [0.002, 0.248]], [[0.002, 0.248], [0.25, 0.00001]]],
    )


def test_call_simple():
    neural_factor, p, q, xor = get_data()

    print("neural_factor({p: 0, q:1, xor: 1}))", neural_factor({p: 0, q: 1, xor: 1}))
    assert neural_factor({p: 0, q: 1, xor: 1}) > 0.9

    print("neural_factor({p: 0, q:1, xor: 0}))", neural_factor({p: 0, q: 1, xor: 0}))
    assert neural_factor({p: 0, q: 1, xor: 0}) < 0.1

    print("neural_factor({p: 1, q:1, xor: 0}))", neural_factor({p: 1, q: 1, xor: 0}))
    assert neural_factor({p: 1, q: 1, xor: 0}) > 0.9


def test_call_multivalue():
    neural_factor, p, q, xor = get_data()

    assignment = {p: torch.tensor([0, 1]), q: 1, xor: 1}
    expected = [0.9, 0.1]
    run_test_call_multivalue(assignment, expected, neural_factor)

    assignment = {p: [0, 1], q: 1, xor: 1}
    expected = [0.9, 0.1]
    run_test_call_multivalue(assignment, expected, neural_factor)

    assignment = {p: [0, 1], q: [1, 1], xor: 1}
    expected = [0.9, 0.1]
    run_test_call_multivalue(assignment, expected, neural_factor)

    assignment = {p: [0, 1], q: [1, 1], xor: [1, 1]}
    expected = [0.9, 0.1]
    run_test_call_multivalue(assignment, expected, neural_factor)

    assignment = {p: [], q: [], xor: 1}
    expected = []
    run_test_call_multivalue(assignment, expected, neural_factor)


def run_test_call_multivalue(assignment, expected, neural_factor):
    expected_tensor = torch.tensor(expected)
    actual = neural_factor(assignment)
    print("neural_factor({p: [0, 1], q:1, xor: 1}))", actual)
    small = torch.tensor([0.05]*len(actual))
    assert all((actual - expected_tensor) ** 2 < small)




import functools
import itertools
import math
import operator
from typing import List

from neuralpp.inference.graphical_model.representation.factor.factor import Factor
from neuralpp.util import util
from neuralpp.util.group import Group
from neuralpp.util.util import join, split


class ProductFactor(Factor):
    def __init__(self, factors: List[Factor]):
        super().__init__(set.union(*(set(f.variables) for f in factors)))
        self._factors = util.flatten_one_level(factors, util.isinstance_predicate(ProductFactor), ProductFactor.factors)

    def call_after_validation(self, assignment_dict, assignment_values):
        return math.prod(f(assignment_dict) for f in self._factors)

    def condition_on_non_empty_dict(self, assignment_dict):
        return ProductFactor(f.condition(assignment_dict) for f in self._factors)

    def randomize(self):
        for f in self._factors:
            f.randomize()

    def randomized_copy(self):
        return ProductFactor(f.randomized_copy() for f in self._factors)

    def __mul__(self, other):
        if isinstance(other, ProductFactor):
            additional = other._factors
        else:
            additional = [other]
        return ProductFactor(self._factors + additional)

    def sum_out_variable(self, variable):
        factors_with_variable, factors_without_variable = split(
            self._factors, lambda f: variable in f
        )
        result_of_summing_out_variable_from_product_of_factors_with_variable = (
            Group.product(factors_with_variable) ^ variable
        )
        if factors_without_variable:
            return ProductFactor(
                factors_without_variable
                + [result_of_summing_out_variable_from_product_of_factors_with_variable]
            )
        else:
            # at this point, result_of_summing_out_variable_from_product_of_factors_with_variable
            # may be a ProductFactor or an atomic factor.
            # Here we ensure it is a ProductFactor
            return ProductFactor.make(
                result_of_summing_out_variable_from_product_of_factors_with_variable
            )

    @staticmethod
    def make(factor):
        return ProductFactor(ProductFactor.factors(factor))

    @staticmethod
    def factors(product_of_factors):
        if isinstance(product_of_factors, ProductFactor):
            return product_of_factors._factors
        else:  # singleton
            return [product_of_factors]

    def atomic_factor(self):
        return Group.product(self._factors)

    def argmax(self):
        return self.atomic_factor().argmax()

    def normalize(self):
        return self.atomic_factor().normalize()

    def __str__(self):
        return join(self._factors, " * ")



from neuralpp.util.util import matrix_from_function


def test_matrix_from_function():

    tests = [
        (([range(4)], lambda i: i), [0, 1, 2, 3]),
        (([range(2), range(3)], lambda i, j: i * 3 + j), [[0, 1, 2], [3, 4, 5]]),
        (
            ([range(2), range(3), range(2)], lambda i, j, k: i * 6 + j * 2 + k),
            [[[0, 1], [2, 3], [4, 5]], [[6, 7], [8, 9], [10, 11]]],
        ),
        (([], lambda: 0), 0),
        (([range(0)], lambda i: i), []),
        (([range(0), range(3)], lambda i, j: i * 3 + j), []),
        (([range(2), range(0), range(2)], lambda i, j, k: i * 6 + j * 2 + k), [[], []]),
    ]

    for test in tests:
        assert matrix_from_function(*test[0]) == test[1]



import math

import torch
from neuralpp.util.timer import Timer


LOG_OF_NEAR_ZERO = -35.0
NEAR_ZERO = 1e-15


def log_of_nested_list_without_inf_non_differentiable(o):
    if isinstance(o, list):
        return [log_of_nested_list_without_inf_non_differentiable(e) for e in o]
    elif isinstance(o, torch.Tensor):
        return log_without_inf_non_differentiable(o)
    else:
        try:
            if o < NEAR_ZERO:
                return LOG_OF_NEAR_ZERO
            else:
                return math.log(o)
        except Exception as e:
            print(
                f"Error in log_of_nested_list_without_inf_non_differentiable for argument {o}: {e}"
            )
            raise e


# def fix_log_of_nested_list(o):
#     if isinstance(o, list):
#         return [fix_log_of_nested_list(e) for e in o]
#     elif isinstance(o, torch.Tensor):
#         return fix_log(o)
#     elif o == -np.inf:
#         return LOG_OF_NEAR_ZERO
#     else:
#         return o


def log_without_inf_non_differentiable(tensor):
    non_zero_tensor = fix_zeros(tensor)
    result = non_zero_tensor.log()
    return result


# def fix_log(tensor):
#     tensor_clone = tensor.clone()
#     indices_of_infs = torch.isinf(tensor)
#     tensor_clone[indices_of_infs] = LOG_OF_NEAR_ZERO
#     return tensor_clone


def fix_zeros(tensor):
    if tensor.shape == ():
        return fix_zeros_in_non_dimensional_tensor(tensor)
    else:
        return fix_zeros_in_non_zero_dimensional_tensor(tensor)


def fix_zeros_in_non_dimensional_tensor(tensor):
    # I expected the code for the dimensional tensor case to work for the non-dimensional case as well,
    # but for some reason PyTorch doesn't work with that, so we do it separately here.
    if tensor == 0:
        return torch.tensor(NEAR_ZERO)
    else:
        return tensor


def fix_zeros_in_non_zero_dimensional_tensor(tensor):
    whether_it_is_zero = tensor == 0
    indices_of_zeros = whether_it_is_zero.nonzero(as_tuple=True)
    if len(indices_of_zeros) != 0:
        tensor_clone = tensor.clone()
        tensor_clone[indices_of_zeros] = NEAR_ZERO
        return tensor_clone
    else:
        return tensor


# def sum_in_log_space(x, y):
#     """
#     Given tensors x, y, computes log(e^x + e^y).
#     If x > y, then this is more accurate than the straightforward definition and also works for y = -np.inf.
#     """
#     return x + torch.log1p((y - x).log)



import random
import time

import torch
from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import (
    GraphicalModelSGDLearner,
)
from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    generalized_len_of_dict_frame,
)
from neuralpp.inference.graphical_model.representation.frame.multi_frame_data_loader import (
    MultiFrameDataLoader,
)
from neuralpp.inference.graphical_model.representation.model.model import (
    compute_accuracy_on_frames_data_loader,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.graphical_model.variable.tensor_variable import TensorVariable
from neuralpp.inference.neural_net.ConvNet import ConvNet
from neuralpp.inference.neural_net.from_log_to_probabilities_adapter import (
    FromLogToProbabilitiesAdapter,
)
from neuralpp.util.generic_sgd_learner import default_after_epoch
from neuralpp.util.mnist_util import read_mnist, show_images_and_labels
from neuralpp.util.pickle_cache import pickle_cache
from neuralpp.util.util import set_default_tensor_type_and_return_device, go_up_until_we_have_subdirectory

show_examples = True
shuffle_data = True
batch_size = 100
deterministic_seed = True
debug = False
max_datapoints = 100000  # all
# max_datapoints = 5000
recompute_dataset_from_scratch = False
try_cuda = True


def main():

    if deterministic_seed:
        torch.manual_seed(0)
        random.seed(0)

    digit_var = IntegerVariable("digit_var", 10)
    image_var = TensorVariable("image_var", non_batch_dim=2)

    # Read dataset before setting default tensor type to cuda

    print("Getting NeuPP dataset ready...")
    go_up_until_we_have_subdirectory("data")
    train_dataset = pickle_cache(
        lambda: make_dataset(digit_var, image_var, "train"),
        "data/cache/simple_mnist_batcheable_train_dataset.pkl",
        refresh=recompute_dataset_from_scratch,
    )

    if show_examples:
        show_first_examples(train_dataset, 2, 3)

    device = set_default_tensor_type_and_return_device(try_cuda)
    print(f"Using {device} device")

    # Create model after setting default tensor type to cuda

    conv_net = FromLogToProbabilitiesAdapter(ConvNet())

    model = [
        NeuralFactor(conv_net, input_variables=[image_var], output_variable=digit_var),
    ]

    print(f"Computing accuracy before training...")
    train_data_loader = MultiFrameDataLoader(train_dataset)
    accuracy = compute_accuracy_on_frames_data_loader(train_data_loader, model, device)
    print(f"Accuracy on training dataset before training: {accuracy*100:.2f}%")

    def epoch_hook(learner):
        default_after_epoch(learner, end_str=" ")
        accuracy = compute_accuracy_on_frames_data_loader(
            learner.data_loader, model, device
        )
        time_elapsed = time.time() - learner.time_start
        print(
            f"[{time_elapsed:.0f} s] Accuracy on training dataset: {accuracy*100:.2f}%"
        )

    print("Learning...")
    GraphicalModelSGDLearner(
        model, train_data_loader, after_epoch=epoch_hook, debug=debug, device=device
    ).learn()


def make_dataset(digit_var, image_var, phase="train"):

    print("Reading MNIST dataset...")
    images_by_digits = read_mnist(max_datapoints)

    digits = range(10)

    all_images_tensor = torch.cat(
        [torch.stack(images_by_digits[phase][d]) for d in digits]
    )
    all_digits_tensor = torch.cat(
        [torch.tensor([d]).repeat(len(images_by_digits[phase][d])) for d in digits]
    )

    if shuffle_data:
        points = len(all_images_tensor)
        permutation = torch.randperm(points)
        all_permuted_images_tensor = all_images_tensor[permutation]
        all_permuted_digits_tensor = all_digits_tensor[permutation]
    else:
        all_permuted_images_tensor = all_images_tensor
        all_permuted_digits_tensor = all_digits_tensor

    dataset = [
        (
            {image_var: all_permuted_images_tensor},
            {digit_var: all_permuted_digits_tensor},
        )
    ]

    return dataset


def show_first_examples(dataset, n_rows, n_cols):
    first_observed_frame, first_query_frame = dataset[0]
    (
        images,
    ) = first_observed_frame.values()  # first_observed_frame is a singleton dict
    (labels,) = first_query_frame.values()  # first_query_frame is a singleton dict
    show_images_and_labels(n_rows, n_cols, images, labels)


if __name__ == "__main__":

    main()



from neuralpp.inference.graphical_model.representation.factor.continuous.continuous_internal_parameterless_factor import \
    ContinuousInternalParameterlessFactor


class PyTorchDistributionFactor(ContinuousInternalParameterlessFactor):
    """
    A specialization of ContinuousInternalParameterlessFactor based on a PyTorch distribution.
    """

    def __init__(self, pytorch_distribution_maker, variables, conditioning_dict=None):
        """
        Makes a PyTorchDistributionFactor based on given PyTorch distribution.
        The first element of 'variables' must be the value following the distribution,
        and the remaining ones must be the distribution parameters in the same
        order used by the distribution maker function.
        """
        super().__init__(variables, conditioning_dict)
        self.pytorch_distribution_maker = pytorch_distribution_maker
        self.value_variable = self.variables[0]
        self.distribution_parameter_variables = self.variables[1:]

    def condition_on_non_empty_dict(self, assignment_dict):
        return type(self)(self.variables, self.total_conditioning_dict(assignment_dict))

    def call_after_validation(self, assignment_dict, assignment_values):
        assignment_and_conditioning_dict = self.total_conditioning_dict(assignment_dict)
        distribution_parameters = [assignment_and_conditioning_dict[variable]
                                   for variable in self.distribution_parameter_variables]
        distribution = self.pytorch_distribution_maker(distribution_parameters)
        value = assignment_and_conditioning_dict[self.value_variable]
        return distribution.log_prob(value).exp()



import time


class Timer:
    def __init__(self, *description_args):
        if description_args and type(description_args[0]) == bool:
            self.active = description_args[0]
            description_args = description_args[1:]
        else:
            self.active = True
        self.description_args = description_args

    def __enter__(self):
        if self.active:
            print("\nStarted", *self.description_args)
            self.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.active:
            self.end = time.time()
            print(
                "Finished",
                *(self.description_args + (f"({self.end - self.start:.2f} seconds)",)),
            )


if __name__ == "__main__":
    with Timer("test"):
        [i for i in range(100000)]
    with Timer("test2"):
        [i for i in range(100000)]
    with Timer(True, "This should appear"):
        [i for i in range(100000)]
    with Timer(False, "This should not appear!"):
        [i for i in range(100000)]



import functools
from typing import Optional, Tuple, Union, Any

import torch

from neuralpp.inference.graphical_model.variable.variable import Variable

FlexibleShape = Union[torch.Size, Tuple[int, ...]]


class TensorVariable(Variable):
    def __init__(self, name, non_batch_dim: int):
        """
        Constructs a tensor variable with given name and optional non-batch shape.

        Non-batch shape is used for deciding whether a value should be considered a batch or not.
        In some applications this is not needed, so this information is optional.
        An exception will be raised if this is not provided but a functionality
        dependent on it is needed.
        """
        super().__init__()
        self.name = name
        self.non_batch_dim = non_batch_dim

    def featurize(self, value) -> torch.Tensor:
        self._check_value_is_tensor(value)
        return value

    def is_multivalue(self, value: Any) -> bool:
        self._check_value_is_tensor(value)
        self._check_value_dimension(value)
        return value.dim() == self.non_batch_dim + 1

    def multivalue_len(self, value: Any) -> int:
        assert self.is_multivalue(value)
        return value.shape[0]

    @staticmethod
    def _check_value_is_tensor(value):
        if not isinstance(value, torch.Tensor):
            raise Exception(f"Values of {TensorVariable.__name__} must be tensors.")

    def _check_value_dimension(self, tensor):
        if tensor.dim() not in {self.non_batch_dim, self.non_batch_dim + 1}:
            raise Exception(f"Tensor value for {self} must have dimension "
                            f"{self.non_batch_dim} or {self.non_batch_dim + 1} (if batch)")

    def __eq__(self, other):
        assert isinstance(
            other, TensorVariable
        ), "TensorVariable being compared to non-TensorVariable"
        result = self.name == other.name
        return result

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return str(self.name)



from types import SimpleNamespace

import torch

from neuralpp.experiments.mnist_semi_supervised.mnist_semi_supervised import default_parameters, \
    solve_learning_problem_from_parameters
from neuralpp.inference.graphical_model.learn.learning_problem_solver import solve_learning_problem, LearningProblem
from neuralpp.inference.graphical_model.learn.uniform_training import UniformTraining
from neuralpp.inference.graphical_model.representation.factor.fixed.fixed_pytorch_factor import (
    FixedPyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_log_table import (
    PyTorchLogTable,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.graphical_model.variable.tensor_variable import TensorVariable
from neuralpp.inference.neural_net.ConvNet import ConvNet
from neuralpp.inference.neural_net.MLP import MLP
from neuralpp.inference.neural_net.from_log_to_probabilities_adapter import (
    FromLogToProbabilitiesAdapter,
)
from neuralpp.util.data_loader_from_random_data_point_thunk import (
    data_loader_from_batch_generator,
)
from neuralpp.util.generic_sgd_learner import default_after_epoch
from neuralpp.util.mnist_util import read_mnist, show_images_and_labels
from neuralpp.util.util import join, set_seed

# Trains a digit recognizer with the following factor graph:
#
#                      Constraint0
#                           |
#        +------------------------------------+
#        |    Constraint0 = Digit0 + Digit1   |
#        +------------------------------------+
#               |                      |
#            Digit0                  Digit1
#               |                      |
#               |                      |
#      +------------------+   +------------------+
#      | Digit recognizer |   | Digit recognizer |
#      +------------------+   +------------------+
#               |                      |
#               |                      |
#             Image0                 Image1


def indices_of_digit_arguments_of_constraint(constraint_index):
    return 0, 1


def constraint_function(constraint_index, d0, d1):
    # Using a single constraint, so constraint_index is irrelevant
    return d0 + d1


def generate_chain_of_summands_digits_values_and_constraint_equal_to_sum(number_of_digits,
        chain_length, # ignored because we have always two digits
        number_of_constraints, # ignored because we have always 1 constraint
        batch_size):
    d0_values = torch.randint(number_of_digits, (batch_size,))
    max_d1_value = number_of_digits - d0_values
    d1_values = torch.cat([torch.randint(max_d1_value[i], (1,)) for i in range(batch_size)])
    constraint_values = d0_values + d1_values
    # print(d0_values[0].item(), d1_values[0].item(), constraint_values[0].item())
    return [d0_values, d1_values], [constraint_values]


parameters = default_parameters()
parameters.chain_length = 2
parameters.number_of_constraints = 1
parameters.number_of_constraint_values = parameters.number_of_digits
parameters.indices_of_digit_arguments_of_constraint = indices_of_digit_arguments_of_constraint
parameters.constraint_function = constraint_function
parameters.number_of_constraints = 1
# parameters.custom_digits_and_constraints_values_batches_generator = None # not available because it generates invalid examples (eg 6 + 7 = 13)
parameters.custom_digits_and_constraints_values_batches_generator = generate_chain_of_summands_digits_values_and_constraint_equal_to_sum

solve_learning_problem_from_parameters(parameters)



class Table:
    def __init__(self):
        self._batch = None
        self._non_batch_shape = None

    @property
    def batch(self):
        return self._batch

    @batch.setter
    def batch(self, value):
        self._batch = value

    @property
    def non_batch_shape(self):
        return self._non_batch_shape

    @non_batch_shape.setter
    def non_batch_shape(self, value):
        self._non_batch_shape = value

    def __getitem__(self, non_batch_slice_coordinates):
        """
        Returns a tensor corresponding to the values indexed by given non-batch coordinates,
        that is, coordinates for dimensions that are not the batch row index dimension.
        A coordinate can be either an integer, slice(None), or an array-like object of integers
        To describe the result, let us divide into cases.
        If there are no multivalue coordinates and the table is not a batch,
        then the result is table[non_batch_slice_coordinates].
        This will be either a scalar, if there are no slice(None) coordinates, or a tensor-like object if there
        are slice(None) coordinates.
        If there are no batch coordinates and the table is a batch, then the result is
        M such that M_i is table[i, *non_batch_slice_coordinates],
        that is, for each row of the table batch, we apply the given non-batch coordinates.
        Note that this will be 1D is there were no slice(None) coordinates, and of greater dimensions otherwise.
        If there are batch coordinates and the table is not a batch, then the result is
        M such that M_i is table[*th(i, non_batch_slice_coordinates)] where
        th is a function that derives a tuple of values from non_batch_slice_coordinates using only the i-th
        element of each batch-coordinate (slice(None) coordinates are left alone).
        Note that this requires all batch coordinates to contain the same number of values.
        Finally, if table is a batch and there are batch coordinates, then the result is
        M such that M_i is table[i, *th(i, non_batch_slice_coordinates)], that is,
        we apply the same procedure for each row i.
        """
        self._not_implemented("__getitem__")

    def __len__(self):
        self._not_implemented("__len__")

    def assignments(self):
        self._not_implemented("assignments")

    def __mul__(self, item):
        self._not_implemented("__mul__")

    def expand(self, shape, dim):
        self._not_implemented("expand")

    def permute(self, permutation):
        self._not_implemented("permute")

    def sum_out(self, dim):
        self._not_implemented("sum_out")

    def sum(self):
        self._not_implemented("sum")

    def argmax(self):
        """
        Returns index arrays (or scalar if there is a single non-batch dimension) for the largest value in the table.
        If the table is a batch, the indices will be a batch of such index arrays or scalars.
        """
        self._not_implemented("argmax")

    def normalize(self):
        self._not_implemented("normalize")

    def sample(self):
        self._not_implemented("single_sample")

    def slice(self, non_batch_slice_coordinates):
        """
        Returns a table corresponding to the data returned from __getitem__ (see its documentation).
        The resulting table will be a batch if it contains 0 or more than 1 rows, and not a batch otherwise.
        """
        self._not_implemented("condition")

    def randomize(self):
        self._not_implemented("randomize")

    def randomized_copy(self):
        self._not_implemented("randomized_copy")

    def _not_implemented(self, name):
        # creating a variable first prevents compiler from thinking this is an abstract method
        error = NotImplementedError(f"{name} not implemented for {type(self)}")
        raise error



class DataLoaderFromEpochGeneratorThunk:
    def __init__(self, generator):
        self.generator = generator

    def __iter__(self):
        return self.generator()


def random_epoch_generator(epoch_size, random_data_point_thunk, print=print):
    for batch_number in range(epoch_size):
        if print is not None and batch_number % 50 == 0:
            print(f"Batch: {batch_number}/{epoch_size}")
        data_point = random_data_point_thunk()
        yield data_point


def data_loader_from_batch_generator(
    epoch_size, random_data_point_generator, print=print
):
    return DataLoaderFromEpochGeneratorThunk(
        lambda: random_epoch_generator(
            epoch_size, random_data_point_generator, print=print
        )
    )



import torch

from neuralpp.util.generic_sgd_learner import GenericSGDLearner


class UniformTraining(GenericSGDLearner):
    def __init__(self, model, data_loader, number_of_classes):
        super().__init__(model, data_loader)
        self.number_of_classes = number_of_classes
        self.singleton_probability = [1.0 / self.number_of_classes]

    def loss_function(self, batch):
        probabilities = self.model(batch)
        uniform_probabilities = torch.tensor(self.singleton_probability).repeat(len(batch), self.number_of_classes)
        loss = torch.square(uniform_probabilities - probabilities).sum()
        return loss


import torch


def shape(variables):
    return tuple(v.cardinality for v in variables)


def insert_shape(original_shape, shape_to_be_inserted, dim):
    """Returns a shape resulting from inserting a given shape into an original shape at position dim"""
    shape_before_dim = original_shape[0:dim]
    shape_after_dim = original_shape[dim:]
    final_shape = shape_before_dim + shape_to_be_inserted + shape_after_dim
    return final_shape


def n_unsqueeze(tensor, n, dim):
    """Unsqueezes (that is, inserts a dimension of size 1 at the specified position dim) n times"""
    current = tensor
    for i in range(n):
        current = torch.unsqueeze(current, dim)
    return current


def index_of(elements, a_list):
    return [a_list.index(element) for element in elements]


def permutation_from_to(seq1, seq2):
    return [seq2.index(e) for e in seq1]



from typing import Any, Dict

from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.variable.discrete_variable import (
    DiscreteVariable,
)
from neuralpp.util import util
from neuralpp.util.util import merge_dicts


def check_and_show_conditional_distributions(
    neural_factor, ground_truth=None, tolerance=0.05
):
    """
    Checks if neural_factor implements a distribution (probabilities summing up to 1 for each input assignment)
    and shows it in the console.
    If a ground truth function is provided, checks the neural factor distribution against it.
    The ground truth is passed an assignment as an unpacked list in the same order as the variables in the factor.
    """
    input_variables = neural_factor.input_variables
    output_variable = neural_factor.output_variable
    for input_assignment_dict in DiscreteVariable.assignments_product_dicts(
        input_variables
    ):

        potentials = [
            compute_potential(
                neural_factor, input_assignment_dict, output_assignment_dict
            )
            for output_assignment_dict in DiscreteVariable.assignments_product_dicts(
                [output_variable]
            )
        ]
        assert abs(sum(potentials) - 1) < 0.0001

        for output_assignment_dict in DiscreteVariable.assignments_product_dicts(
            [output_variable]
        ):
            potential, assignment_dict = compute_potential_and_assignment(
                neural_factor, input_assignment_dict, output_assignment_dict
            )
            print("----------------------")
            print(
                f"Neural factor potential for {assignment_dict}:",
                util.rounded_list(potential),
                end="",
            )

            if ground_truth is not None:
                assignment = [assignment_dict[v] for v in neural_factor.variables]
                ground_probability = ground_truth(*assignment)
                print(f" (ground truth: {round(ground_probability, 3)})")
                if not util.close(potential, ground_probability, tolerance):
                    raise AssertionError(
                        f"Probability {round(potential.item(), 3)} not close enough to "
                        f"ground truth {ground_probability} (tolerance is {tolerance})"
                    )
            else:
                print()


def compute_potential(
    neural_factor: NeuralFactor,
    input_assignment_dict: Dict[DiscreteVariable, Any],
    output_assignment_dict: Dict[DiscreteVariable, Any],
):
    potential, assignment_dict = compute_potential_and_assignment(
        neural_factor, input_assignment_dict, output_assignment_dict
    )
    return potential


def compute_potential_and_assignment(
    neural_factor: NeuralFactor,
    input_assignment_dict: Dict[DiscreteVariable, Any],
    output_assignment_dict: Dict[DiscreteVariable, Any],
):
    assignment_dict = merge_dicts(input_assignment_dict, output_assignment_dict)
    potential = neural_factor(assignment_dict)
    return potential, assignment_dict



import random

import pytest
import torch
from neuralpp.inference.graphical_model.representation.table.pytorch_log_table import (
    PyTorchLogTable,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_table import (
    BatchCoordinatesDoNotAgreeException,
    PyTorchTable,
)
from neuralpp.util import util
from neuralpp.util.util import check_that_exception_is_thrown, matrix_from_function


@pytest.fixture(params=["Non-log space", "Log space"])
def class_to_use(request):
    log_space = request.param == "Log space"
    if log_space:
        class_to_use = PyTorchLogTable
    else:
        class_to_use = PyTorchTable
    return class_to_use


@pytest.fixture(params=["No batch", "Empty batch", "Batch"])
def batch_size(request):
    return (
        -1
        if request.param == "No batch"
        else 0
        if request.param == "Empty batch"
        else 10
    )


@pytest.fixture(params=["No batch 1", "Empty batch 1", "Batch 1"])
def batch_size1(request):
    return (
        -1
        if request.param == "No batch 1"
        else 0
        if request.param == "Empty batch 1"
        else 10
    )


@pytest.fixture(params=["No batch 2", "Empty batch 2", "Batch 2"])
def batch_size2(request):
    return (
        -1
        if request.param == "No batch 2"
        else 0
        if request.param == "Empty batch 2"
        else 10
    )


@pytest.fixture(params=["No batch empty", "Empty batch empty", "Batch empty"])
def batch_size_empty(request):
    return (
        -1
        if request.param == "No batch empty"
        else 0
        if request.param == "Empty batch empty"
        else 10
    )


def f1_xy(x, y):
    return float(x * 2 + y)


def f1_ixy(i, x, y):
    return 10 ** i * f1_xy(x, y)


@pytest.fixture
def table1(class_to_use, batch_size1):

    if batch_size1 == -1:
        table1 = class_to_use.from_function((3, 2), (range(3), range(2)), f1_xy)
    else:
        table1 = class_to_use.from_function(
            (batch_size1, 3, 2),
            (range(batch_size1), range(3), range(2)),
            f1_ixy,
            batch=True,
        )
    print(f"table1: {table1}")
    return table1


@pytest.fixture
def table2(class_to_use, batch_size2):
    if batch_size2 == -1:
        table2 = class_to_use.from_array([0.7, 0.2, 0.1])
    else:
        table2 = class_to_use.from_function(
            (batch_size2, 3),
            (range(batch_size2), range(3)),
            lambda i, j: [0.7, 0.2, 0.1][j],
            batch=True,
        )
    print(f"table2: {table2}")
    return table2


@pytest.fixture
def empty(class_to_use, batch_size_empty):
    if batch_size_empty == -1:
        empty = class_to_use.from_array([])
    else:
        empty = class_to_use.from_function(
            (batch_size_empty, 0),
            (range(batch_size_empty), range(0)),
            lambda i: None,
            batch=True,
        )
    print(f"empty: {empty}")
    return empty


def test_init_table1(table1, batch_size1):
    print(f"table1: {table1}")
    if batch_size1 == -1:
        table1_data = matrix_from_function([range(3), range(2)], f1_xy)
    else:
        table1_data = matrix_from_function(
            [range(batch_size1), range(3), range(2)], f1_ixy
        )
    table1_shape = (3, 2)
    table1_length = 6
    run_init_test(batch_size1, table1, table1_length, table1_data, table1_shape)


def test_init_table2(table2, batch_size2):
    print(f"table2: {table2}")
    table2_data = [0.7, 0.2, 0.1] * (1 if batch_size2 == -1 else batch_size2)
    table2_shape = (3,)
    table2_length = 3
    run_init_test(batch_size2, table2, table2_length, table2_data, table2_shape)


def test_init_empty(empty, batch_size_empty):
    print(f"empty: {empty}")
    empty_data = []
    empty_shape = (0,)
    empty_length = 0
    run_init_test(batch_size_empty, empty, empty_length, empty_data, empty_shape)


def run_init_test(batch_size, table, table_length, table_data, table_shape):
    if batch_size == -1:
        assert torch.allclose(table.potentials_tensor(), torch.tensor(table_data))
        assert len(table) == table_length
    else:
        assert torch.allclose(
            table.potentials_tensor(),
            torch.tensor(table_data).reshape(batch_size, *(table_shape)),
        )
        assert len(table) == table_length * batch_size


def test_assignments1(table1):
    # batch does not change the set of assignments
    print(f"table1 assignments: {list(table1.assignments())}")
    assert list(table1.assignments()) == [
        (0, 0),
        (0, 1),
        (1, 0),
        (1, 1),
        (2, 0),
        (2, 1),
    ]


def test_assignments2(table2):
    # batch does not change the set of assignments
    print(f"table2 assignments: {list(table2.assignments())}")
    assert list(table2.assignments()) == [(0,), (1,), (2,)]


def test_assignments_empty(empty):
    # batch does not change the set of assignments
    print(f"empty assignments: {list(empty.assignments())}")
    assert list(empty.assignments()) == []


def test_expand1(class_to_use, table1, batch_size1):
    run_expand_test(table1, batch_size1, class_to_use)


def test_expand2(class_to_use, table2, batch_size2):
    run_expand_test(table2, batch_size2, class_to_use)


def test_expand_empty(class_to_use, empty, batch_size_empty):
    run_expand_test(empty, batch_size_empty, class_to_use)


def run_expand_test(table, batch_size, class_to_use):
    for i in range(10):
        shape_to_insert_length = random.randrange(5)
        shape_to_insert = tuple(
            random.randrange(5) for i in range(shape_to_insert_length)
        )
        dim = random.randrange(len(table.non_batch_shape))
        expected_table_expand = make_expected_expanded(
            table, shape_to_insert, dim, class_to_use, batch_size
        )
        actual_table_expand = table.expand(shape_to_insert, dim)
        assert actual_table_expand == expected_table_expand


def make_expected_expanded(table, shape_to_insert, dim, class_to_use, batch_size):
    if batch_size == -1:
        batch_dim = ()
    else:
        batch_dim = (batch_size,)
    new_shape = (
        batch_dim
        + table.non_batch_shape[:dim]
        + shape_to_insert
        + table.non_batch_shape[dim:]
    )
    new_ranges = [range(d) for d in new_shape]

    def f(*args):
        effective_dim = dim if batch_size == -1 else dim + 1
        args_without_expansion = (
            args[:effective_dim] + args[effective_dim + len(shape_to_insert) :]
        )
        return table.potentials_tensor()[args_without_expansion].item()
        # Note that we need to use table.potentials_tensor()[] rather than table[]
        # because the latter only takes non-batch coordinates into account but here we need to use the batch index.

    expected_table_expand = class_to_use.from_function(
        new_shape, new_ranges, f, batch=(batch_size != -1)
    )
    return expected_table_expand


def test_permute1(class_to_use, table1, batch_size1):
    run_permute_test(table1, batch_size1, class_to_use)


def test_permute2(class_to_use, table2, batch_size2):
    run_permute_test(table2, batch_size2, class_to_use)


def test_permute_empty(class_to_use, empty, batch_size_empty):
    run_permute_test(empty, batch_size_empty, class_to_use)


def run_permute_test(table, batch_size, class_to_use):
    for i in range(10):
        non_batch_number_of_dimensions = len(table.non_batch_shape)
        permutation = random.sample(
            range(non_batch_number_of_dimensions), non_batch_number_of_dimensions
        )
        expected_table_permutation = make_expected_permutation(
            table, permutation, class_to_use, batch_size
        )
        actual_table_permutation = table.permute(permutation)
        assert actual_table_permutation == expected_table_permutation


def make_expected_permutation(table, permutation, class_to_use, batch_size):
    if batch_size == -1:
        effective_permutation = permutation
    else:
        effective_permutation = [0] + [p + 1 for p in permutation]

    inverse_permutation = compute_inverse_permutation(effective_permutation)
    potentials_tensor = table.potentials_tensor()

    def f(*args_in_permuted_order):
        args_in_original_order = tuple(
            args_in_permuted_order[i] for i in inverse_permutation
        )
        return potentials_tensor[args_in_original_order]

    permuted_shape = permute(table.shape(), effective_permutation)
    permuted_ranges = [range(i) for i in permuted_shape]

    expected_table_permutation = class_to_use.from_function(
        permuted_shape, permuted_ranges, f, batch=(batch_size != -1)
    )
    return expected_table_permutation


def permute(array, permutation):
    return [array[i] for i in permutation]


def compute_inverse_permutation(permutation):
    result = permutation.copy()
    for i, v in enumerate(permutation):
        result[v] = i
    return result


def test_non_batch_slice1(class_to_use, table1, batch_size1):
    """
    Tests slice without batch coordinates (multiple values for the same coordinate).
    """

    test_coordinates = [
        (slice(None), slice(None)),
        (slice(None), 1),
        (1, slice(None)),
        (1, 1),
    ]

    run_non_batch_slice_tests_no_assumption_of_equal_batch_rows(
        table1,
        f1_ixy,
        table1.non_batch_shape,
        batch_size1,
        class_to_use,
        test_coordinates,
    )


def run_non_batch_slice_tests_no_assumption_of_equal_batch_rows(
    table,
    function,
    original_non_batch_shape,
    batch_size,
    class_to_use,
    test_coordinates,
):

    for coordinates in test_coordinates:
        number_of_data_rows = 1 if batch_size == -1 else batch_size
        expected_data = []

        def process_from_dim(dim, prefix_arguments):
            if dim == len(original_non_batch_shape):
                expected_data.append(function(*prefix_arguments))
            else:
                for v in range(original_non_batch_shape[dim]):
                    if coordinates[dim] == slice(None) or coordinates[dim] == v:
                        process_from_dim(dim + 1, prefix_arguments + [v])

        for row in range(number_of_data_rows):
            process_from_dim(0, prefix_arguments=[row])

        non_batch_shape = tuple(
            original_non_batch_shape[dim]
            for dim, _ in enumerate(original_non_batch_shape)
            if coordinates[dim] == slice(None)
        )

        if batch_size == -1:
            shape = tuple(non_batch_shape)
        else:
            shape = (number_of_data_rows, *non_batch_shape)
        expected_tensor = torch.tensor(expected_data).reshape(shape)
        expected = class_to_use.from_array(expected_tensor, batch=(batch_size != -1))
        assert table.slice(coordinates) == expected
        assert torch.allclose(table[coordinates], expected_tensor)


def test_non_batch_slice2(class_to_use, table2, batch_size2):
    """
    Tests slice without batch coordinates (multiple values for the same coordinate).
    """
    tests = [
        ((slice(None),), [0.7, 0.2, 0.1]),
        ((slice(0, 2),), [0.7, 0.2]),
        ((1,), 0.2),
    ]

    run_non_batch_slice_tests_assuming_equal_batch_rows(
        table2, "table2", tests, batch_size2, class_to_use
    )


def test_non_batch_slice_empty(class_to_use, empty, batch_size_empty):
    """
    Tests slice without batch coordinates (multiple values for the same coordinate).
    """
    tests = [
        ((), []),
    ]

    run_non_batch_slice_tests_assuming_equal_batch_rows(
        empty, "empty", tests, batch_size_empty, class_to_use
    )


def run_non_batch_slice_tests_assuming_equal_batch_rows(
    table, table_name, tests, batch_size, class_to_use
):
    """
    Tests slice without batch coordinates (multiple values for the same coordinate).
    """
    for coordinates, expected_row_array in tests:
        actual = table.slice(coordinates)
        print(f"{table_name}.slice(): {actual}")
        expected = make_table_from_row_array(
            class_to_use, batch_size, expected_row_array
        )
        print("expected:", expected)
        assert actual == expected


def run_non_batch_slice_tests2(table, table_name, tests, batch_size, class_to_use):
    """
    Tests slice without batch coordinates (multiple values for the same coordinate).
    """
    for coordinates, expected_non_batch_data_array in tests:
        actual = table.slice(coordinates)
        print(f"{table_name}.slice(): {actual}")
        expected = make_table_from_non_batch_data_array(
            class_to_use, batch_size, expected_non_batch_data_array
        )
        print("expected:", expected)
        assert actual == expected


def test_batch_slice1(class_to_use, table1, batch_size1):

    # Checking valid cases

    if batch_size1 == -1:
        coordinates = [1, 1]
        expected = class_to_use.from_function(
            shape=tuple(),
            function_arguments_iterables=tuple(),
            function_of_potentials=lambda: table1[1, 1],
            batch=False,
        )
        actual = table1.slice(coordinates)
        assert actual == expected
        for n in [0, 1, 10]:
            coordinates = [[1] * n, [1] * n]
            expected = class_to_use.from_function(
                shape=(n,),
                function_arguments_iterables=(range(n),),
                function_of_potentials=lambda row: f1_xy(1, 1),
                batch=True,
            )
            actual = table1.slice(coordinates)
            assert actual == expected
    elif batch_size1 == 0:
        for coordinates in [
            [1, []],
            [[], 1],
            [[], []],
        ]:
            expected = class_to_use.from_function(
                shape=(
                    0,
                ),  # only row dimension returns since non-batch dimensions would be determined.
                function_arguments_iterables=(range(0),),
                function_of_potentials=lambda row: None,
                batch=True,
            )
            actual = table1.slice(coordinates)
            assert actual == expected
    else:  # multi-row batch
        n = table1.number_of_batch_rows()
        assert n % 2 == 0, "This test assumes the batch has an even number of rows"
        for coordinates in [
            [1, 1],
            [[0, 1] * (n // 2), [0, 1] * (n // 2)],
            [1, [1] * n],
            [[1] * n, 1],
        ]:
            at_row = lambda row, c: c[row] if isinstance(c, list) else c
            expected = class_to_use.from_function(
                shape=(
                    n,
                ),  # only row dimension returns since non-batch dimensions would be determined.
                function_arguments_iterables=(range(n),),
                function_of_potentials=lambda i: f1_ixy(
                    i, at_row(i, coordinates[0]), at_row(i, coordinates[1])
                ),
                batch=True,
            )
            actual = table1.slice(coordinates)
            assert actual == expected

    # Checking invalid cases

    if batch_size1 == -1:
        coordinates = [
            [0, 1, 0],
            [0, 1],
        ]  # batch coordinates with difference numbers of values
        check_that_exception_is_thrown(
            lambda: table1.slice(coordinates), BatchCoordinatesDoNotAgreeException
        )
    elif batch_size1 == 0:
        coordinates = [1, [0, 1]]
        # batch coordinates (batch itself and second coordinate) with difference numbers of values
        check_that_exception_is_thrown(
            lambda: table1.slice(coordinates), BatchCoordinatesDoNotAgreeException
        )
    else:  # multi-row batch
        n = table1.number_of_batch_rows()
        coordinates = [
            [1] * (n // 2),
            [1] * n,
        ]  # coordinates not agreeing on number of values
        check_that_exception_is_thrown(
            lambda: table1.slice(coordinates), BatchCoordinatesDoNotAgreeException
        )
        coordinates = [
            [1] * (n // 2),
            [1] * (n // 2),
        ]  # different from neuralpp.number of batch rows
        check_that_exception_is_thrown(
            lambda: table1.slice(coordinates), BatchCoordinatesDoNotAgreeException
        )


def test_get_item2(class_to_use, table2, batch_size2):
    tests = [
        ((slice(None),), [0.7, 0.2, 0.1]),
        ((slice(0, 2),), [0.7, 0.2]),
        ((1,), 0.2),
    ]

    run_get_item_test(tests, table2, batch_size2)


def run_get_item_test(tests, table, batch_size):
    for coordinates, expected_row_array in tests:
        actual = table[coordinates]
        print(f"{coordinates}: {actual}")
        expected = make_tensor_from_row_array(batch_size, expected_row_array)
        print("expected:", expected)
        assert actual.shape == expected.shape and torch.allclose(actual, expected)


def test_mul(class_to_use):
    # It is hard to compute the expected tables without basically duplicating the multiplication code,
    # so instead here we follow the more traditional path of testing some key cases.

    # First we test some basic non-batch, non-empty cases

    table1 = class_to_use.from_array([[1.0, 2.0], [3.0, 4.0]], batch=False)
    table2 = class_to_use.from_array([[10.0, 20.0], [30.0, 40.0]], batch=False)
    expected = class_to_use.from_array([[10.0, 40.0], [90.0, 160.0]], batch=False)
    run_mul_test(expected, table1, table2)

    table1 = class_to_use.from_array([[10.0, 20.0], [30.0, 40.0]], batch=False)
    table2 = class_to_use.from_array([[1.0, 2.0], [3.0, 4.0]], batch=False)
    expected = class_to_use.from_array([[10.0, 40.0], [90.0, 160.0]], batch=False)
    run_mul_test(expected, table1, table2)

    # Now we move to some one-batch cases

    table1 = class_to_use.from_array([1.0, 2.0], batch=False)
    table2 = class_to_use.from_array([[10.0, 20.0], [30.0, 40.0]], batch=True)
    expected = class_to_use.from_array([[10.0, 40.0], [30.0, 80.0]], batch=True)
    run_mul_test(expected, table1, table2)

    table1 = class_to_use.from_array([[1.0, 2.0], [3.0, 4.0]], batch=True)
    table2 = class_to_use.from_array([10.0, 20.0], batch=False)
    expected = class_to_use.from_array([[10.0, 40.0], [30.0, 80.0]], batch=True)
    run_mul_test(expected, table1, table2)

    # We continue on to one-batch cases, but now the batches are empty
    # we use tensors to give tables the right shape.

    table1 = class_to_use.from_array([1.0, 2.0], batch=False)
    table2 = class_to_use.from_array(torch.tensor([]).reshape(0, 2), batch=True)
    expected = table2
    run_mul_test(expected, table1, table2)

    table1 = class_to_use.from_array(torch.tensor([]).reshape(0, 2), batch=True)
    table2 = class_to_use.from_array([10.0, 20.0], batch=False)
    expected = table1
    run_mul_test(expected, table1, table2)

    # Now we make even the non-batch side empty as well.

    table1 = class_to_use.from_array([], batch=False)
    table2 = class_to_use.from_array(torch.tensor([]).reshape(0, 0), batch=True)
    expected = table2
    run_mul_test(expected, table1, table2)

    table1 = class_to_use.from_array(torch.tensor([]).reshape(0, 0), batch=True)
    table2 = class_to_use.from_array([], batch=False)
    expected = table1
    run_mul_test(expected, table1, table2)

    # Now we keep the non-batch empty, but make the batches have length greater than zero
    # (still empty thought to match the non-batches).

    table1 = class_to_use.from_array([], batch=False)
    table2 = class_to_use.from_array(torch.tensor([]).reshape(10, 0), batch=True)
    expected = table2
    run_mul_test(expected, table1, table2)

    table1 = class_to_use.from_array(torch.tensor([]).reshape(10, 0), batch=True)
    table2 = class_to_use.from_array([], batch=False)
    expected = table1
    run_mul_test(expected, table1, table2)

    # And now both are empty non-batches.

    table1 = class_to_use.from_array([])
    table2 = table1
    expected = table1
    run_mul_test(expected, table1, table2)

    # Now we enter test with batches on both sizes, of varying sizes.

    # First, both batches have multiple rows

    table1 = class_to_use.from_array([[1.0, 2.0], [3.0, 5.0]], batch=True)
    table2 = class_to_use.from_array([[10.0, 20.0], [30.0, 40.0]], batch=True)
    expected = class_to_use.from_array([[10.0, 40.0], [90.0, 200.0]], batch=True)
    run_mul_test(expected, table1, table2)

    table1 = class_to_use.from_array([[10.0, 20.0], [30.0, 40.0]], batch=True)
    table2 = class_to_use.from_array([[1.0, 2.0], [3.0, 5.0]], batch=True)
    expected = class_to_use.from_array([[10.0, 40.0], [90.0, 200.0]], batch=True)
    run_mul_test(expected, table1, table2)

    # Now, both batches have zero rows

    table1 = class_to_use.from_array(torch.tensor([]).reshape(0, 2), batch=True)
    table2 = class_to_use.from_array(torch.tensor([]).reshape(0, 2), batch=True)
    expected = class_to_use.from_array(torch.tensor([]).reshape(0, 2), batch=True)
    run_mul_test(expected, table1, table2)

    # Now, both are batches but one has zero rows and the other has multiple rows
    # This should fail as the two batches must have the same number of rows
    try:
        table1 = class_to_use.from_array(torch.tensor([]).reshape(0, 2), batch=True)
        table2 = class_to_use.from_array([[1.0, 2.0], [3.0, 5.0]], batch=True)
        actual = table1 * table2
        raise AssertionError(
            "Multiplication of two batches with different number of rows should have but did not."
        )
    except AssertionError:
        pass

    # This should also fail for the same reason
    try:
        table1 = class_to_use.from_array([[1.0, 2.0]], batch=True)
        table2 = class_to_use.from_array([[1.0, 2.0], [3.0, 5.0]], batch=True)
        expected = None
        actual = table1 * table2
        raise AssertionError(
            "Multiplication of two batches with different number of rows should have but did not."
        )
    except AssertionError:
        pass


def run_mul_test(expected, table1, table2):
    actual = table1 * table2
    print("table1:", table1)
    print("table2:", table2)
    print("table1*table2:", actual)
    print("expected:", expected)
    assert actual == expected


def test_sum_out1(class_to_use, table1, batch_size1):

    if batch_size1 == -1:
        actual = table1.sum_out(0)
        expected = class_to_use.from_array([6.0, 9.0], batch=False)
        assert actual == expected

        actual = table1.sum_out(1)
        expected = class_to_use.from_array([1.0, 5.0, 9.0], batch=False)
        assert actual == expected

        actual = table1.sum_out((0, 1))
        expected = class_to_use.from_array(15.0, batch=False)
        assert actual == expected

        actual = table1.sum_out([])
        expected = class_to_use.from_array(15.0, batch=False)
        assert actual == expected

    elif batch_size1 == 0:
        actual = table1.sum_out(0)
        expected = class_to_use.from_array(torch.zeros(0, 2), batch=True)
        assert actual == expected

        actual = table1.sum_out(1)
        expected = class_to_use.from_array(torch.zeros(0, 3), batch=True)
        assert actual == expected

        actual = table1.sum_out((0, 1))
        expected = class_to_use.from_array(torch.zeros(0), batch=True)
        assert actual == expected

        actual = table1.sum_out(())
        expected = class_to_use.from_array(torch.zeros(0), batch=True)
        assert actual == expected

    elif batch_size1 > 0:
        actual = table1.sum_out(0)
        expected = class_to_use.from_array(
            [[6.0 * 10 ** i, 9.0 * 10 ** i] for i in range(batch_size1)], batch=True
        )
        assert actual == expected

        actual = table1.sum_out(1)
        expected = class_to_use.from_array(
            [[1.0 * 10 ** i, 5.0 * 10 ** i, 9.0 * 10 ** i] for i in range(batch_size1)],
            batch=True,
        )
        assert actual == expected

        actual = table1.sum_out((0, 1))
        expected = class_to_use.from_array(
            [15.0 * 10 ** i for i in range(batch_size1)], batch=True
        )
        assert actual == expected

        actual = table1.sum_out([])
        expected = class_to_use.from_array(
            [15.0 * 10 ** i for i in range(batch_size1)], batch=True
        )
        assert actual == expected


def test_sum_out2(class_to_use, table2, batch_size2):
    tests = [
        (0, 1.0),
        # the following does not seem consistent with the previous ones; I would expect the original tensor
        # to be preserved (no dimensions being summed out, so nothing changes), but PyTorch
        # interprets an empty list of dims to mean "all dims".
        ([], 1.0),
        ((), 1.0),
    ]

    run_sum_out_test(table2, "table2", batch_size2, class_to_use, tests)


def test_sum_out_empty(class_to_use, empty, batch_size_empty):
    tests = [
        ([], 0.0),
        ((), 0.0),
    ]

    run_sum_out_test(empty, "empty", batch_size_empty, class_to_use, tests)


def run_sum_out_test(table, table_name, batch_size, class_to_use, tests):
    for dim, expected_array in tests:
        expected = make_table_from_row_array(class_to_use, batch_size, expected_array)
        actual = table.sum_out(dim)
        print(f"{table_name}.sum_out({dim}): {actual}")
        print("expected:", expected)
        assert actual == expected


def test_sum1(class_to_use, table1, batch_size1):
    if batch_size1 == -1:
        actual = table1.sum()
        expected = torch.tensor(15.0)
        assert torch.allclose(actual, expected)

    elif batch_size1 == 0:
        actual = table1.sum()
        expected = torch.zeros(0)
        assert torch.allclose(actual, expected)

    elif batch_size1 > 0:
        actual = table1.sum()
        expected = torch.tensor([15.0 * 10 ** i for i in range(batch_size1)])
        assert torch.allclose(actual, expected)


def test_sum2(class_to_use, table2, batch_size2):
    run_sum_test(table2, "table2", batch_size2, expected_value=1.0)


def test_sum_empty(class_to_use, empty, batch_size_empty):
    run_sum_test(empty, "empty", batch_size_empty, expected_value=0.0)


def run_sum_test(table, table_name, batch_size1, expected_value):
    expected = make_tensor_from_row_array(batch_size1, expected_value)
    actual = table.sum()
    print(f"{table_name}:", actual)
    print("expected:", expected)
    assert actual.shape == expected.shape and torch.allclose(actual, expected)


def test_normalize1(class_to_use, table1, batch_size1):
    if batch_size1 == -1:
        actual = table1.normalize()
        print(actual)
        expected = class_to_use.from_array(
            [[0.0, 0.0667], [0.1333, 0.2], [0.2667, 0.3333]], batch=False
        )
        assert actual == expected

    elif batch_size1 == 0:
        actual = table1.normalize()
        print(actual)
        expected = class_to_use.from_array(torch.zeros(0, 3, 2), batch=True)
        assert actual == expected

    elif batch_size1 > 0:
        actual = table1.normalize()
        print(actual)
        expected = class_to_use.from_array(
            [[[0.0, 0.0667], [0.1333, 0.2], [0.2667, 0.3333]]] * batch_size1, batch=True
        )
        assert actual == expected


def test_normalize2(class_to_use, table2, batch_size2):
    # in this test we use tensors to represent expected rows as opposed to arrays like in the other test
    # because we need to divide all elements by the partition function, an operation that is not available for arrays.
    expected_row_array = (torch.tensor([0.7, 0.2, 0.1]) / 1).tolist()

    run_normalize_test(table2, "table2", batch_size2, expected_row_array, class_to_use)


def test_normalize_empty(class_to_use, empty, batch_size_empty):
    pass  # can't normalize the empty table


def run_normalize_test(table, table_name, batch_size, expected_row_array, class_to_use):
    expected = make_table_from_row_array(class_to_use, batch_size, expected_row_array)
    actual = table.normalize()
    print(f"{table_name}: {actual}")
    print("expected:", expected)
    assert actual == expected


# ###################### auxiliary functions


def close(a, b):
    return abs(a - b) < 0.0001


def make_table_from_row_array(class_to_use, batch_size, expected_row_array):
    batch = batch_size != -1
    if batch:
        # here we must create the table from neuralpp.a tensor which, unlike arrays, is able to represent a shape
        # even if the batch is empty (creating a table from neuralpp.an empty array renders its shape equal to (0)).
        non_batch_shape = util.array_shape(expected_row_array)
        total_shape = (batch_size, *non_batch_shape)
        expected_tensor = torch.tensor([expected_row_array] * batch_size).reshape(
            total_shape
        )
        expected = class_to_use.from_array(expected_tensor, batch)
    else:
        expected = class_to_use.from_array(expected_row_array, batch)
    return expected


def make_table_from_non_batch_data_array(
    class_to_use, batch_size, expected_non_batch_data_array
):
    batch = batch_size != -1
    if batch:
        non_batch_shape = util.array_shape(expected_non_batch_data_array)
        total_shape = (batch_size, *non_batch_shape)
        expected_tensor = torch.tensor(
            [expected_non_batch_data_array] * batch_size
        ).reshape(total_shape)
        expected = class_to_use.from_array(expected_tensor, batch)
    else:
        expected = class_to_use.from_array(expected_non_batch_data_array, batch)
    return expected


def make_tensor_from_row_array(batch_size, expected_row_array):
    batch = batch_size != -1
    if batch:
        non_batch_shape = util.array_shape(expected_row_array)
        total_shape = (batch_size, *non_batch_shape)
        expected = torch.tensor([expected_row_array] * batch_size).reshape(total_shape)
    else:
        expected = torch.tensor(expected_row_array)
    return expected



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# SPDX-License-Identifier: MIT

import sys

from setuptools import find_packages, setup

REQUIRED_MAJOR = 3
REQUIRED_MINOR = 8


TEST_REQUIRES = [
    "cudatoolkit",
    "pytorch",
    "torchaudio",
    "torchvision",
    "matplotlib",
    "tqdm",
    "pytest",
    "pytest-cov"
]
DEV_REQUIRES = TEST_REQUIRES + [
    "black",
    "flake8",
    "flake8-bugbear",
    "mypy",
    "usort",
]


# Check for python version
if sys.version_info < (REQUIRED_MAJOR, REQUIRED_MINOR):
    error = (
        "Your version of python ({major}.{minor}) is too old. You need "
        "python >= {required_major}.{required_minor}."
    ).format(
        major=sys.version_info.major,
        minor=sys.version_info.minor,
        required_minor=REQUIRED_MINOR,
        required_major=REQUIRED_MAJOR,
    )
    sys.exit(error)


# read in README.md as the long description
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="neuralpp",
    version="0.1.0",
    description="Neural Probabilistic Programs (NeuralPPs)",
    author="Rodrigo de Salvo Braz",
    license="MIT",
    url="https://github.com/rodrigodesalvobraz/neuralpp",
    project_urls={
        "Documentation": "https://github.com/rodrigodesalvobraz/neuralpp",
        "Source": "https://github.com/rodrigodesalvobraz/neuralpp",
    },
    keywords=[
        "Bayesian Inference",
        "Exact Inference",
        "PyTorch",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">={}.{}".format(REQUIRED_MAJOR, REQUIRED_MINOR),
    install_requires=[
        "torch>=1.8.1",
    ],
    packages=find_packages(
        include=["neuralpp", "neuralpp.*"],
        exclude=["debug", "tests", "website"],
    ),
    extras_require={
        "dev": DEV_REQUIRES,
        "test": TEST_REQUIRES,
    },
)



import torch


class TensorMixedRadix:
    """
    Converter from integers to their mixed radix representation.
    """

    def __init__(self, radices):
        self.strides, self.max_value = self.compute_strides_and_max_value(radices)

    def representation(self, values: torch.Tensor):
        """
        Given a 1-dimensional tensor of values, returns a 2-dimensional tensor
        where each row i is a tensor with the digits of the representation of values[i]
        in a mixed radix representation with given radices.
        """

        number_of_values = values.shape[0]

        if number_of_values == 0:
            return torch.zeros(0, len(self.strides))

        if (m := max(values)) > self.max_value:
            raise MaxValueException(m, self.max_value)

        representation = torch.zeros((number_of_values, 0), dtype=torch.int)
        for i in range(len(self.strides)):
            i_th_digits = torch.tensor(values // self.strides[i], dtype=torch.int)
            representation = torch.cat(
                (representation, i_th_digits.reshape(number_of_values, 1)), dim=1
            )
            values -= i_th_digits * self.strides[i]
        return representation

    @staticmethod
    def compute_strides_and_max_value(radices):
        if len(radices) == 0:
            return [], 0
        strides = [1] * len(radices)
        for i in range(len(radices) - 2, -1, -1):
            # i in len(radices) - 2, ..., 0
            strides[i] = radices[i + 1] * strides[i + 1]
        max_value = radices[0] * strides[0] - 1
        print(f"strides: {strides}")
        return strides, max_value


class MaxValueException(Exception):
    def __init__(self, violating_value, max_value):
        super().__init__(
            f"Received value {violating_value} greater than mixed radix max value {max_value}"
        )
        self.violating_value = violating_value
        self.max_value = max_value

    def __eq__(self, other):
        return (
            isinstance(other, MaxValueException)
            and self.violating_value == other.violating_value
            and self.max_value == other.max_value
        )



import random

import pytest
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable


@pytest.fixture(params=["Non-log space", "Log space"])
def log_space(request):
    return request.param == "Log space"


@pytest.fixture(params=["No batch", "Empty batch", "Batch"])
def batch_size(request):
    return (
        -1
        if request.param == "No batch"
        else 0
        if request.param == "Empty batch"
        else 10
    )


@pytest.fixture
def x():
    return IntegerVariable("x", 3)


@pytest.fixture
def y():
    return IntegerVariable("y", 2)


@pytest.fixture
def factor1(x, y, log_space):
    # noinspection PyShadowingNames
    result = PyTorchTableFactor.from_function(
        [x, y], lambda x, y: 0.9 if x == y else 0.1, log_space=log_space
    )
    print(f"factor1: {result}")
    return result


@pytest.fixture
def factor2(x, y, log_space):
    # noinspection PyShadowingNames
    result = PyTorchTableFactor([x], [0.7, 0.2, 0.1], log_space=log_space)
    print(f"factor1: {result}")
    return result


@pytest.fixture
def normalized_product_factor(factor1, factor2):
    print("Product:", factor1 * factor2)
    normalized_product = (factor1 * factor2).normalize()
    print("Normalized product:", normalized_product)
    return normalized_product


def test_sample(normalized_product_factor):

    random.seed(1)

    print("Samples from neuralpp.normalized product:")
    actual = [normalized_product_factor.single_sample() for i in range(20)]
    expected = [
        (0, 0),
        (1, 1),
        (1, 0),
        (0, 0),
        (0, 0),
        (0, 0),
        (0, 0),
        (1, 0),
        (0, 0),
        (0, 0),
        (1, 1),
        (0, 0),
        (1, 0),
        (0, 0),
        (0, 0),
        (1, 0),
        (0, 0),
        (1, 1),
        (1, 1),
        (0, 0),
    ]
    print(f"Expected: {expected}")
    print(f"Actual  : {actual}")
    assert actual == expected


def test_normalization(x, y, factor1, factor2):
    product = factor1 * factor2
    print("Product:", product)
    normalized_product = product.normalize()
    print("Normalized product:", normalized_product)
    expected = PyTorchTableFactor(
        [y, x], [[0.6848, 0.0217, 0.0109], [0.0761, 0.1957, 0.0109]]
    )
    assert normalized_product == expected


def test_condition(x, y, factor1, factor2):
    conditioned_nothing = factor1.condition({})
    print(f"factor1 | nothing: {conditioned_nothing}")
    expected = factor1
    assert conditioned_nothing == expected

    conditioned_x_0 = factor1.condition({x: 0})
    print(f"factor1 | x = 0: {conditioned_x_0}")
    expected = PyTorchTableFactor([y], [0.9, 0.1])
    assert conditioned_x_0 == expected

    conditioned_y_0 = factor1.condition({y: 0})
    print(f"factor1 | y = 0: {conditioned_y_0}")
    expected = PyTorchTableFactor([x], [0.9, 0.1, 0.1])
    assert conditioned_y_0 == expected

    conditioned_x_0_y_1 = factor1.condition({x: 0, y: 1})
    print(f"factor1 | x = 0, y = 1: {conditioned_x_0_y_1}")
    expected = PyTorchTableFactor([], 0.1)
    assert conditioned_x_0_y_1 == expected


def test_sum_out(x, y, factor1, factor2):
    print(f"factor1 ^ x: {factor1 ^ x}")
    print(f"factor1 ^ y: {factor1 ^ y}")
    print(f"factor1 ^ [x, y]: {factor1 ^ [x, y]}")
    print(f"factor2 ^ x: {factor2 ^ x}")
    assert PyTorchTableFactor([y], [1.1, 1.1]) == factor1 ^ x
    assert PyTorchTableFactor([x], [1.0, 1.0, 0.2]) == factor1 ^ y
    assert PyTorchTableFactor([], 2.2) == factor1 ^ [x, y]
    assert PyTorchTableFactor([], 1.0) == factor2 ^ x


def test_zeros():
    x = IntegerVariable("x", 2)
    y = IntegerVariable("y", 2)
    for log_space in {True, False}:
        f = PyTorchTableFactor.from_function(
            [x, y], lambda x, y: float(x == y), log_space=log_space
        )
        assert close(f({x: 0, y: 0}), 1.0)
        assert close(f({x: 0, y: 1}), 0.0)
        assert close(f({x: 1, y: 0}), 0.0)
        assert close(f({x: 1, y: 1}), 1.0)


def close(a, b):
    return abs(a - b) < 0.0001



from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import GraphicalModelSGDLearner
from neuralpp.util.util import set_default_tensor_type_and_return_device, join

# This module provides a generic function for solving learning problems.

# It provides a LearningProblem interface for users to extend with the details of their
# problems.
# They can then invoke function solve_learning_problem on their learning problem
# and have it solved.
# See more details in the documentation of each element.

class LearningProblem:
    """
    An interface for learning problems.
    It must know how to provide a data loader, set up a model, evaluate the model,
    and print summaries after each learning epoch.
    """

    def make_data_loader(self):
        """
        Provides a data loader for the problem.
        The data loader must be an iterable over data points or batches
        that can be provided to the self.model as arguments.
        """
        raise NotImplementedError("make_data_loader")

    def setup_model(self):
        """Sets up self.model."""
        raise NotImplementedError("setup_model")

    def learning_is_needed(self):
        """
        Indicates whether learning is actually needed
        (must return False if self.model is already created
        in final form).
        """
        raise NotImplementedError("learning_is_needed")

    def print_evaluation(self):
        """Evaluates self.model, printing a summary."""
        raise NotImplementedError("print_evaluation")

    @property
    def after_epoch(self):
        """See documentation for GraphicalModelSGDLearner after_epoch parameter."""
        raise NotImplementedError("after_epoch")


def solve_learning_problem(
        problem: LearningProblem,
        try_cuda,
        lr,
        loss_decrease_tol,
        max_epochs_to_go_before_stopping_due_to_loss_decrease
):
    """
    Solves a LearningProblem.
    Runs an initial evaluation before learning starts, and another at the end.
    Uses GraphicalModelSGDLearner, passing it problem.model, data loader
    and after_epoch provided by the learning problem.
    """

    # Load data, if needed, before setting default device to cuda
    train_data_loader = problem.make_data_loader()

    device = set_default_tensor_type_and_return_device(try_cuda)

    # Creating model after attempting to set default tensor type to cuda so it sits there
    problem.setup_model()

    print("\nInitial evaluation:")
    problem.print_evaluation()

    if problem.learning_is_needed():
        print("Learning...")
        learner = GraphicalModelSGDLearner(
            problem.model,
            train_data_loader,
            device=device,
            lr=lr,
            loss_decrease_tol=loss_decrease_tol,
            max_epochs_to_go_before_stopping_due_to_loss_decrease=max_epochs_to_go_before_stopping_due_to_loss_decrease,
            after_epoch=problem.after_epoch,
        )
        learner.learn()

    print("\nFinal model:")
    print(join(problem.model, "\n"))

    print("\nFinal evaluation:")
    problem.print_evaluation()



import pytest
import torch

from neuralpp.util.util import repeat_first_dimension_with_expand, repeat_interleave_first_dimension, \
    RepeatFirstDimensionException


def test_repeat_first_dimension_with_expand():
    with pytest.raises(RepeatFirstDimensionException):
        tensor = [1,2,3]
        n = 2
        expected = None
        run_repeat_first_dimension_with_expand(tensor, n, expected)

    with pytest.raises(RepeatFirstDimensionException):
        tensor = torch.tensor(1)
        n = 2
        expected = None
        run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([1,2,3])
    n = 2
    expected = torch.tensor([1, 2, 3, 1, 2, 3])
    run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([[1,2,3], [4,5,6]])
    n = 2
    expected = torch.tensor([[1,2,3], [4,5,6], [1,2,3], [4,5,6]])
    run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([])
    n = 2
    expected = torch.zeros(0, dtype=torch.float)
    run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([])
    n = 0
    expected = torch.zeros(0, dtype=torch.float)
    run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([[1,2,3], [4,5,6]])
    n = 0
    expected = torch.zeros(0, 3, dtype=torch.long)
    run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([
        [[1,2,3],
         [4,5,6]],

        [[10,20,30],
         [40,50,60]],
    ])
    n = 2
    expected = torch.tensor([
        [[1, 2, 3],
         [4, 5, 6]],

        [[10, 20, 30],
         [40, 50, 60]],

        [[1, 2, 3],
         [4, 5, 6]],

        [[10, 20, 30],
         [40, 50, 60]],
    ])
    run_repeat_first_dimension_with_expand(tensor, n, expected)

    tensor = torch.tensor([
        [[1,2,3],
         [4,5,6]],

        [[10,20,30],
         [40,50,60]],
    ])
    n = 0
    expected = torch.zeros(0, 2, 3, dtype=torch.long)
    run_repeat_first_dimension_with_expand(tensor, n, expected)


def run_repeat_first_dimension_with_expand(tensor, n, expected):
    actual = repeat_first_dimension_with_expand(tensor, n)
    assert actual.equal(expected)


def test_repeat_interleave_first_dimension():
    with pytest.raises(RepeatFirstDimensionException):
        tensor = [1,2,3]
        n = 2
        expected = None
        run_repeat_interleave_first_dimension(tensor, n, expected)

    with pytest.raises(RepeatFirstDimensionException):
        tensor = torch.tensor(1)
        n = 2
        expected = None
        run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([1,2,3])
    n = 2
    expected = torch.tensor([1,1,2,2,3,3])
    run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([[1,2,3], [4,5,6]])
    n = 2
    expected = torch.tensor([[1,2,3], [1,2,3], [4,5,6], [4,5,6]])
    run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([])
    n = 2
    expected = torch.zeros(0, dtype=torch.float)
    run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([])
    n = 0
    expected = torch.zeros(0, dtype=torch.float)
    run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([[1, 2, 3], [4, 5, 6]])
    n = 0
    expected = torch.zeros(0, 3, dtype=torch.long)
    run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([
        [[1,2,3],
         [4,5,6]],

        [[10,20,30],
         [40,50,60]],
    ])
    n = 2
    expected = torch.tensor([
        [[1, 2, 3],
         [4, 5, 6]],

        [[1, 2, 3],
         [4, 5, 6]],

        [[10, 20, 30],
         [40, 50, 60]],

        [[10, 20, 30],
         [40, 50, 60]],
    ])
    run_repeat_interleave_first_dimension(tensor, n, expected)

    tensor = torch.tensor([
        [[1,2,3],
         [4,5,6]],

        [[10,20,30],
         [40,50,60]],
    ])
    n = 0
    expected = torch.zeros(0, 2, 3, dtype=torch.long)
    run_repeat_interleave_first_dimension(tensor, n, expected)


def run_repeat_interleave_first_dimension(tensor, n, expected):
    actual = repeat_interleave_first_dimension(tensor, n)
    assert actual.equal(expected)



import math
from collections import Counter

import pytest
import torch
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import PyTorchTableFactor
from neuralpp.inference.graphical_model.representation.table.pytorch_log_table import (
    PyTorchLogTable,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_table import (
    BatchCoordinatesDoNotAgreeException,
    PyTorchTable,
)
from neuralpp.inference.graphical_model.variable.discrete_variable import DiscreteVariable
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.util.discrete_sampling import discrete_sample


@pytest.fixture
def x():
    return IntegerVariable("x", 3)


@pytest.fixture
def y():
    return IntegerVariable("y", 2)


@pytest.fixture
def z():
    return IntegerVariable("z", 2)


@pytest.fixture(params=["Non-log space", "Log space"])
def log_space(request):
    return request.param == "Log space"


@pytest.fixture(params=["Non-log space", "Log space"])
def log_space1(request):
    return request.param == "Log space"


@pytest.fixture(params=["Non-log space", "Log space"])
def log_space2(request):
    return request.param == "Log space"


@pytest.fixture(params=["Non-log space", "Log space"])
def log_space_expected(request):
    return request.param == "Log space"


@pytest.fixture(params=[None, 0, 4])
def batch_size(request):
    return request.param


@pytest.fixture(params=[None, 0, 4])
def batch_size1(request):
    return request.param


@pytest.fixture(params=[None, 0, 4])
def batch_size2(request):
    return request.param


def test_assignments(x, y, log_space):
    factor = PyTorchTableFactor.from_function(
        (x, y), lambda x, y: float(x == y), log_space=log_space
    )
    print(list(factor.assignments()))
    assert list(factor.assignments()) == [
        (0, 0),
        (0, 1),
        (1, 0),
        (1, 1),
        (2, 0),
        (2, 1),
    ]


def test_new_instance(x, y, log_space):
    factor1 = PyTorchTableFactor.from_function(
        (x, y), lambda x, y: float(x == y), log_space=log_space
    )
    factor2 = factor1.new_instance(factor1.variables, factor1.table)
    assert factor2 is not factor1
    assert factor2 == factor2


def batch_function_adapter_from_function_without_batch_row_index(
    function_without_batch_row_index, batch
):
    def result(*args):
        args_without_batch_row_index = args[1:] if batch else args
        potential_without_batch_row_index = function_without_batch_row_index(
            *args_without_batch_row_index
        )
        if batch:
            batch_row_index = args[0]
            potential = (batch_row_index + 1) * potential_without_batch_row_index
        else:
            potential = potential_without_batch_row_index
        return potential

    return result


def batch_function_adapter_from_function_with_batch_row_index(
    function_with_batch_row_index, batch
):
    if batch:
        return function_with_batch_row_index
    else:
        return lambda *rest: function_with_batch_row_index(0, *rest)


def table_factor_from_function_without_batch_row_index(
    variables, function_without_batch_row_index, log_space, batch_size
):
    batch = batch_size is not None
    function = batch_function_adapter_from_function_without_batch_row_index(
        function_without_batch_row_index, batch
    )
    factor = PyTorchTableFactor.from_function(
        variables, function, log_space=log_space, batch_size=batch_size
    )
    return factor


def table_factor_from_function_with_batch_row_index(
    variables, function_with_batch_row_index, log_space, batch_size
):
    batch = batch_size is not None
    function = batch_function_adapter_from_function_with_batch_row_index(
        function_with_batch_row_index, batch
    )
    factor = PyTorchTableFactor.from_function(
        variables, function, log_space=log_space, batch_size=batch_size
    )
    return factor


def test_condition(x, y, log_space, log_space_expected, batch_size):
    def expected_table_factor(variables, function):
        return table_factor_from_function_with_batch_row_index(
            variables, function, log_space_expected, batch_size
        )

    fixy = lambda i, x, y: float((10 ** i) * (x * 3 + y))

    factor = table_factor_from_function_with_batch_row_index(
        (x, y), fixy, log_space, batch_size
    )

    table_class_to_use = PyTorchLogTable if log_space_expected else PyTorchTable

    tests = [
        ({}, factor),
        ({x: 0}, expected_table_factor((y,), lambda i, y: fixy(i, 0, y))),
        ({x: 1}, expected_table_factor((y,), lambda i, y: fixy(i, 1, y))),
        ({y: 0}, expected_table_factor((x,), lambda i, x: fixy(i, x, 0))),
        ({y: 1}, expected_table_factor((x,), lambda i, x: fixy(i, x, 1))),
        ({x: 0, y: 0}, expected_table_factor(tuple(), lambda i: fixy(i, 0, 0))),
        ({x: 1, y: 0}, expected_table_factor(tuple(), lambda i: fixy(i, 1, 0))),
        (
            {x: slice(None), y: 0},
            expected_table_factor((x,), lambda i, x: fixy(i, x, 0)),
        ),
        (
            {x: slice(None), y: 1},
            expected_table_factor((x,), lambda i, x: fixy(i, x, 1)),
        ),
        (
            {x: 0, y: slice(None)},
            expected_table_factor((y,), lambda i, y: fixy(i, 0, y)),
        ),
        (
            {x: 1, y: slice(None)},
            expected_table_factor((y,), lambda i, y: fixy(i, 1, y)),
        ),
        ({x: slice(None), y: slice(None)}, expected_table_factor((x, y), fixy)),
    ]

    run_condition_tests(factor, tests)

    tests_for_batch_size_different_from_zero = [
        (
            {x: [0, 2, 2, 2], y: 0},
            PyTorchTableFactor(
                tuple(),
                [fixy(0, 0, 0), fixy(0, 2, 0), fixy(0, 2, 0), fixy(0, 2, 0)],
                log_space=log_space_expected,
                batch=True,
            )
            if batch_size is None
            else PyTorchTableFactor(
                tuple(),
                [fixy(0, 0, 0), fixy(1, 2, 0), fixy(2, 2, 0), fixy(3, 2, 0)],
                log_space=log_space_expected,
                batch=True,
            )
            if batch_size == 4
            else None,
        ),
        (
            {x: [0, 2, 2, 2], y: [0, 1, 0, 1]},
            PyTorchTableFactor(
                tuple(),
                [fixy(0, 0, 0), fixy(0, 2, 1), fixy(0, 2, 0), fixy(0, 2, 1)],
                log_space=log_space_expected,
                batch=True,
            )
            if batch_size is None
            else PyTorchTableFactor(
                tuple(),
                [fixy(0, 0, 0), fixy(1, 2, 1), fixy(2, 2, 0), fixy(3, 2, 1)],
                log_space=log_space_expected,
                batch=True,
            )
            if batch_size == 4
            else None,
        ),
        (
            {x: 2, y: [0, 1, 0, 1]},
            PyTorchTableFactor(
                tuple(),
                [fixy(0, 2, 0), fixy(0, 2, 1), fixy(0, 2, 0), fixy(0, 2, 1)],
                log_space=log_space_expected,
                batch=True,
            )
            if batch_size is None
            else PyTorchTableFactor(
                tuple(),
                [fixy(0, 2, 0), fixy(1, 2, 1), fixy(2, 2, 0), fixy(3, 2, 1)],
                log_space=log_space_expected,
                batch=True,
            )
            if batch_size == 4
            else None,
        ),
    ]

    if batch_size is None or batch_size != 0:
        run_condition_tests(factor, tests_for_batch_size_different_from_zero)

    if batch_size is not None:
        illegal_conditionings_for_batches = [
            {
                x: [1] * (batch_size + 2),
                y: [1] * (batch_size + 2),
            },  # does not coincide with batch size
            # note that using batch_size + 1 would result in [1] coordinates for batch_size == 0,
            # which have length 1 and are therefore *not* considered batch_coordinates, so that would not be illegal!
            # We use batch_size + 2 to get an illegal case for both batch_size == 0 and batch_size == 4.
            {x: [0, 1], y: [0, 1, 0]},  # batch coordinates do not coincide
        ]

        for illegal_conditioning_for_batch in illegal_conditionings_for_batches:
            try:
                factor[illegal_conditioning_for_batch]
                raise AssertionError(
                    f"Should have thrown a {BatchCoordinatesDoNotAgreeException.__name__}"
                )
            except BatchCoordinatesDoNotAgreeException:
                pass

    if batch_size is None:
        illegal_conditionings_for_non_batches = [
            {x: [0, 1], y: [0, 1, 0]},  # batch coordinates do not coincide
        ]

        for illegal_conditioning_for_non_batch in illegal_conditionings_for_non_batches:
            try:
                factor[illegal_conditioning_for_non_batch]
                raise AssertionError(
                    f"Should have thrown a {BatchCoordinatesDoNotAgreeException.__name__}"
                )
            except BatchCoordinatesDoNotAgreeException:
                pass


def run_condition_tests(factor, tests):
    for assignment_dict, expected_factor in tests:
        actual_factor = factor.condition(assignment_dict)
        print(f"factor.condition({assignment_dict}) = {actual_factor}")
        assert actual_factor == expected_factor


def test_get_item(x, y, log_space, log_space_expected, batch_size):
    factor = table_factor_from_function_without_batch_row_index(
        (x, y), lambda x, y: float(x == y), log_space, batch_size
    )

    actual = factor[{x: 0, y: 0}]
    print(f"actual: {actual}")

    expected_value_0_0 = 1.0

    batch = batch_size is not None

    if batch:
        # (batch row index + 1) * expected_value_0_0
        expected = torch.tensor(
            [
                expected_value_0_0 * (batch_row_index + 1)
                for batch_row_index in range(batch_size)
            ]
        )
    else:
        expected = torch.tensor(expected_value_0_0)

    print(f"expected: {expected}")
    assert torch.allclose(actual, expected)


def test_mul(x, y, z, log_space1, log_space2, batch_size):

    f_x_y = lambda x, y: float((x + 1) * (y + 1))
    f_y_z = lambda y, z: float((y + 1) * (z + 1) * 10)

    factor1 = table_factor_from_function_without_batch_row_index(
        (x, y), f_x_y, log_space1, batch_size
    )
    factor2 = table_factor_from_function_without_batch_row_index(
        (y, z), f_y_z, log_space2, batch_size
    )

    batch = batch_size is not None

    def f_x_y_z(*args):
        if batch:
            batch_row_index, x, y, z = args
            batch_aware_f_x_y = (batch_row_index + 1) * f_x_y(x, y)
            batch_aware_f_y_z = (batch_row_index + 1) * f_y_z(y, z)
            result = batch_aware_f_x_y * batch_aware_f_y_z
            print()
            print(f"batch_row_index: {batch_row_index}")
            print(f"batch_row_index + 1: {(batch_row_index + 1)}")
            print(f"x: {x}")
            print(f"y: {y}")
            print(f"z: {z}")
            print(f"f(x,y): {f_x_y(x, y)}")
            print(f"f(y,z): {f_y_z(y, z)}")
            print(
                f"(batch_row_index + 1) * f_x_y(x,y) * (batch_row_index + 1) * f_y_z(y,z): {result}"
            )
        else:
            x, y, z = args
            result = f_x_y(x, y) * f_y_z(y, z)
            print()
            print(f"x: {x}")
            print(f"y: {y}")
            print(f"z: {z}")
            print(f"f(x,y): {f_x_y(x,y)}")
            print(f"f(y,z): {f_y_z(y,z)}")
            print(f"f_x_y(x,y) * f_y_z(y,z): {result}")

        return result

    expected_product = PyTorchTableFactor.from_function(
        (x, y, z), f_x_y_z, log_space_expected, batch_size
    )

    product = factor1 * factor2

    print(f"factor1: {factor1}")
    print(f"factor2: {factor2}")
    print(f"factor1*factor2 : {product}")
    print(f"expected product: {expected_product}")

    assert product == expected_product


def get_assignment_index(assignment, variables):
    current_stride = 1
    total = 0
    for i, v in reversed(list(enumerate(variables))):
        total += assignment[i] * current_stride
        current_stride *= v.cardinality
    return total


def test_get_assignment_index(x, y, z):
    variables = [x, y, z]
    for i, v in enumerate(variables):
        selected_variables = variables[:i]
        assignment_index = 0
        for assignment in DiscreteVariable.assignments_product(selected_variables):
            assert assignment_index == get_assignment_index(
                assignment, selected_variables
            )
            assignment_index += 1


def test_argmax(x, y, z, log_space, batch_size):
    def fixyz(i, x, y, z):
        return sum(v * 10 ** (j + 1) for j, v in enumerate([i, x, y, z]))

    factor = table_factor_from_function_with_batch_row_index(
        (x, y, z), fixyz, log_space, batch_size
    )
    if batch_size is None:
        expected = {x: 2, y: 1, z: 1}
    else:
        expected = {
            x: torch.tensor([2] * batch_size),
            y: torch.tensor([1] * batch_size),
            z: torch.tensor([1] * batch_size),
        }
    actual = factor.argmax()
    for v in (x, y, z):
        assert actual[v].eq(expected[v]).all()


def test_sample(x, y, z, log_space, batch_size):

    for number_of_variables in range(1, 3):
        variables = [x, y, z][0:number_of_variables]

        cardinalities = [v.cardinality for v in variables]

        number_of_assignments = math.prod(v.cardinality for v in variables)
        potentials = list(range(number_of_assignments))
        probabilities = torch.tensor(potentials, dtype=torch.float) / sum(potentials)

        if batch_size is None:

            def f(*assignment):
                return get_assignment_index(assignment, variables)

        else:

            def f(*assignment):
                batch_index, assignment = assignment[0], assignment[1:]
                return get_assignment_index(assignment, variables)

        factor = PyTorchTableFactor.from_function(
            variables, f, log_space=log_space, batch_size=batch_size
        ).normalize()
        # TODO: we should be able to sample from neuralpp.unnormalized factors

        n = 10000

        batch_samples = [factor.sample() for i in range(n)]
        effective_batch_size = batch_size if batch_size is not None else 1
        samples_per_factor_batch_row = [
            [tuple(batch_samples[j][i].tolist()) for j in range(n)]
            for i in range(effective_batch_size)
        ]
        for batch_row in range(effective_batch_size):
            samples_for_row = samples_per_factor_batch_row[batch_row]
            count = Counter(samples_for_row)
            counts_in_order_of_assignment_index = [
                count[a] for a in factor.assignments()
            ]
            z = sum(counts_in_order_of_assignment_index)
            empirical_probabilities = (
                torch.tensor(counts_in_order_of_assignment_index, dtype=torch.float) / z
            )

            # Let X_i be the random indicator that the batch_row-th assignment is sampled.
            # The mean of X_i is the probability p_i that the batch_row-th assignment is sampled.
            # The standard deviation of X_i is std_i = sqrt(p_i * (1 - p_i)).
            # The empirical probability e_i of X_i is the mean of n samples_for_row of X_i.
            # It has, by the CLT, distribution N(p_i, std_i/sqrt(n)).
            # std_i/sqrt(n) is the standard error std_err_i.
            # The probability that e_i is within p_i +- 3.5 std_err_i is practically one.
            # Therefore the probability that e_i is within p_i +- max_i(std_err_i) is also practically one.

            # We compute the maximum standard error by iteration over all p_i.
            # However, for number_of_assignments >=4 the last probability will always have the largest standard error
            # because all p_i <= 5 and they are monotonically increasing and if you plot the std_dev_o(p) graph,
            # it is a parabola with vertex at p = 0.5.
            if number_of_assignments < 4:
                max_std_err = max(std_err(p, n) for p in probabilities)
            else:
                max_std_err = std_err(probabilities[-1], n)

            number_of_standard_errors = (
                10  # TODO: should be 3.5; waiting to figuring that out
            )
            absolute_tolerance = number_of_standard_errors * max_std_err
            assert torch.allclose(
                probabilities, empirical_probabilities, atol=absolute_tolerance
            ), "Samples deviated from neuralpp.expected distribution; this is possible but should be an extremely rare event."


def std_err(p, n):
    return p * (1 - p) / math.sqrt(n)


def test_torch_categorical_sampling():
    def bernoulli_std_err(p, n):
        return p * (1 - p) / math.sqrt(n)

    masses = [0, 1, 2, 3, 4, 5]
    probabilities = torch.tensor(masses, dtype=torch.float) / sum(masses)
    n = 10000
    std_errors = torch.tensor([bernoulli_std_err(p, n) for p in probabilities])
    max_std_error = max(std_errors)
    pretty_much_100_percent_range = 10 * max_std_error

    categorical = torch.distributions.Categorical(probabilities)
    pytorch_sampling_function = lambda: categorical.sample()

    def python_sampling_function():
        return discrete_sample(range(6), lambda i: probabilities[i])

    for trials in range(10):
        run_clt_test(
            python_sampling_function, probabilities, n, pretty_much_100_percent_range
        )

    for trials in range(10):
        run_clt_test(
            pytorch_sampling_function, probabilities, n, pretty_much_100_percent_range
        )


def run_clt_test(sampling_function, probabilities, n, pretty_much_100_percent_range):
    count = get_sampling_count_per_value(n, sampling_function)
    empirical_probabilities = torch.tensor(count, dtype=torch.float) / n
    print(f"     Real probabilities: {probabilities}")
    print(f"Empirical probabilities: {empirical_probabilities}")
    violation = torch.where(
        abs(empirical_probabilities - probabilities) > pretty_much_100_percent_range
    )[0]
    if len(violation) > 0:
        print(f"Violation at index {violation}")
        print(
            f"Empirical is {empirical_probabilities[violation]}, real is {probabilities[violation]}"
        )
        print(
            f"Difference is {abs(empirical_probabilities[violation] - probabilities[violation])}"
        )
        print(f"100% range is {pretty_much_100_percent_range:.10f}")
        raise Exception("CLT apparently violated")


def get_sampling_count_per_value(n, sampling_function):
    count = [0] * 6
    for i in range(n):
        sampled_value = sampling_function()
        count[sampled_value] += 1
    return count



from types import SimpleNamespace

import torch

from neuralpp.inference.graphical_model.learn.learning_problem_solver import solve_learning_problem, LearningProblem
from neuralpp.inference.graphical_model.learn.uniform_training import UniformTraining
from neuralpp.inference.graphical_model.representation.factor.fixed.fixed_pytorch_factor import (
    FixedPyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_log_table import (
    PyTorchLogTable,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.graphical_model.variable.tensor_variable import TensorVariable
from neuralpp.inference.neural_net.ConvNet import ConvNet
from neuralpp.inference.neural_net.MLP import MLP
from neuralpp.inference.neural_net.from_log_to_probabilities_adapter import (
    FromLogToProbabilitiesAdapter,
)
from neuralpp.util.data_loader_from_random_data_point_thunk import (
    data_loader_from_batch_generator,
)
from neuralpp.util.generic_sgd_learner import default_after_epoch
from neuralpp.util.mnist_util import read_mnist, show_images_and_labels
from neuralpp.util.util import join, set_seed


# Trains a digit recognizer with the following factor graph:
#
#                         Constraint0                                           Constraint_k              ...
#                              |                                                     |
#                   +---------------------+                               +---------------------+
#                   |   Constraint0 <=>   |                 ...           |   Constraint_k <=>  |
#                   |  0-th condition     |                               |    <k-th condition> |         ...
#                   +---------------------+                               +---------------------+
#                             |||  (different constraints involve different digits) |||
#     -------------------------------------------------------------------------------------------------------
#               |                                            |
#               |                                            |
#             Digit0                                       Digit1                                        ...
#               |                                            |
#               |                                            |
#      +------------------+                         +------------------+
#      | Digit recognizer |                         | Digit recognizer |
#      +------------------+                         +------------------+
#               |                                            |
#               |                                            |
#             Image0                                       Image1
#
# that is to say, n variables for images, n for digits, (k + 1) for constraints,
# where constraints are defined (each of them) on a particular subset of digit variables,
# according to user-provided functions.
#
# Note that while in initial versions constraints were constrained to be boolean,
# their value is generalized to be any Python value.
#
# This script offers multiple ways of running the above, including a simplified setting
# in which the "images" are just integers from 0 to 9.
# Even in this very simple setting, the recognizer still needs to learn to associate each digit to itself.
# This provides a way of testing the general dynamic of the model without dealing with the complexities
# of actual image recognition.
#
# Other options include whether the two recognizers are the same network or not
# (essential for learning from positive examples only),
# whether negative examples are present (non-consecutive digits with Constraint = false),
# whether to use a single digit image per digit,
# and various possible initializations for the recognizer.


# -------------- DEFAULT PARAMETERS

def default_parameters():
    number_of_digits = 10

    use_real_images = True  # use real images; otherwise, use its digit value only as input (simpler version of experiment)
    use_conv_net = False  # if use_real_images, neural net used is ConvNet for MNIST; otherwise, a simpler MLP for MNIST.
    custom_digits_and_constraints_values_batches_generator = None
    # Custom examples generator provided by user; if None, use random sequences of digits.

    show_examples = False  # show some examples of images (sanity check for data structure)
    use_a_single_image_per_digit = (
        False  # to make the problem easier -- removes digit variability from the problem
    )
    try_cuda = True
    batch_size = 200
    number_of_batches_between_updates = (
        1  # batches have two functions: how much to fit into CUDA if using it, and
    )
    # how many examples to observe before updating.
    # Here we are splitting those two functions, leaving batch_size for the
    # number of datapoints processed at a time, but allowing for updating
    # only after a number of batches are processed.
    # This allows for a better estimation of gradients at each update,
    # decreasing the influence of random fluctuations in these estimates for
    # each update, but may make learning much slower
    number_of_batches_per_epoch = 100
    number_of_epochs_between_evaluations = 1
    max_real_mnist_datapoints = None
    number_of_digit_instances_in_evaluation = 1000
    seed = None  # use None for non-deterministic seed
    evaluation_probability_precision = 2

    use_shared_recognizer = (
        True  # the same recognizer is shared by both "image" -> digit pairs
    )
    # Note that, if given positive examples only, we need a shared recognizer for
    # learning to have a chance of succeeding
    # because otherwise the two recognizers never get to see 9 and 0 respectively.

    # 'recognizer_type' below allows the selection of different functions and initializations for recognizer.
    # If use_real_images is True, this selection is ignored and a ConvNet or MLP is used, along with number_of_digits = 10.

    recognizer_type = "neural net"  # if use_real_images, a ConvNet or MLP, depending on option use_conv_net
    # if use_real_images == False, an MLP with a single input unit, number_of_digits hidden units,
    # and number_of_digits output units.

    # recognizer_type = "fixed ground truth table"  # the correct table, with fixed parameters.

    # recognizer_type = "random table"  # an initialization with random potentials; see parameter below for increasing randomness

    # recognizer_type = "noisy left-shift"  # a hard starting point, in which digits map to their left-shift.
    # This will get most consecutive pairs to satisfy the constraint
    # but for (0, 1) and (8,9), even though *every* digit is being misclassified.
    # See parameter below for selecting noise level.
    # Making this noisier helps because it "dillutes" the hard starting point.
    left_shift_noise = (
        0.1  # probability of noisy left-shift recognizer initialization not left-shifting
        # but hitting some other digit uniformly
    )

    recognizer_type = "uniform"  # a uniform table -- learning works well

    upper_bound_for_log_potential_in_random_table = (
        1  # log of potentials are uniformly sampled from
    )
    # [0, upper_bound_for_log_potential_in_random_table].
    # The higher the value, the farther from the uniform the table is.
    # So far we observe that tables farther from the uniform
    # often get stuck in local minima,
    # and that uniform tables always converge to the correct answer.

    lr = 1e-3 if use_real_images else 1e-3
    loss_decrease_tol = lr * 0.0001

    max_epochs_to_go_before_stopping_due_to_loss_decrease = 15 if batch_size * number_of_batches_between_updates < 500 else 1

    return SimpleNamespace(**locals())


# -------------- END OF DEFAULT PARAMETERS


def make_digits_and_all_true_constraints_values_batches_generator(digit_values_batch_generator):
    """
    A convenience method for taking a generator of digits values batches
    and returning a generator of digit values batches and all-true constraints (value 1).
    This is useful for problems that have boolean constraints and we have
    a generator for digits values that is guaranteed to generate examples satisfying
    all constraints.
    """
    def generate(number_of_digits,
        chain_length,
        number_of_constraints,
        batch_size):

        digit_values = digit_values_batch_generator(
            number_of_digits, chain_length, batch_size)
        constraint_values = [  # TODO: is returning a tensor more efficient?
            torch.ones(batch_size).long()
            for _ in range(number_of_constraints)
        ]
        return digit_values, constraint_values

    return generate


class MNISTChainsProblem(LearningProblem):

    def __init__(
            self,
            number_of_digits,
            chain_length,
            number_of_constraints,
            number_of_constraint_values,

            indices_of_digit_arguments_of_constraint,
            constraint_function,
            custom_digits_and_constraints_values_batches_generator,

            use_real_images,
            use_conv_net,

            show_examples,
            use_a_single_image_per_digit,
            batch_size,
            number_of_batches_between_updates,
            number_of_batches_per_epoch,
            number_of_epochs_between_evaluations,
            max_real_mnist_datapoints,
            number_of_digit_instances_in_evaluation,
            seed,
            evaluation_probability_precision,

            use_shared_recognizer,
            recognizer_type,
            left_shift_noise,
            upper_bound_for_log_potential_in_random_table,
    ):
        self.number_of_digits = number_of_digits
        self.chain_length = chain_length
        self.number_of_constraints = number_of_constraints
        self.number_of_constraint_values = number_of_constraint_values

        self.indices_of_digit_arguments_of_constraint = indices_of_digit_arguments_of_constraint
        self.constraint_function = constraint_function
        self.custom_digits_and_constraints_values_batches_generator = custom_digits_and_constraints_values_batches_generator

        self.use_real_images = use_real_images
        self.use_conv_net = use_conv_net

        self.show_examples = show_examples
        self.use_a_single_image_per_digit = use_a_single_image_per_digit
        self.batch_size = batch_size
        self.number_of_batches_between_updates = number_of_batches_between_updates
        self.number_of_batches_per_epoch = number_of_batches_per_epoch
        self.number_of_epochs_between_evaluations = number_of_epochs_between_evaluations
        self.max_real_mnist_datapoints = max_real_mnist_datapoints
        self.number_of_digit_instances_in_evaluation = number_of_digit_instances_in_evaluation
        self.seed = seed
        self.evaluation_probability_precision = evaluation_probability_precision

        self.use_shared_recognizer = use_shared_recognizer
        self.recognizer_type = recognizer_type
        self.left_shift_noise = left_shift_noise
        self.upper_bound_for_log_potential_in_random_table = upper_bound_for_log_potential_in_random_table

        # to be defined later:

        # data generation attributes
        self.image_random_variables = None
        self.digit_variables = None
        self.constraint_variables = None
        self.images_by_digits_by_phase = None
        self.get_image_batch_from_digit_batch = None
        self.next_image_index_by_digit = None

        # model attributes
        self.model = None
        self.recognizer_factors = None
        self.constraint_factors = None

        # Adjustments:
        if self.use_real_images:
            if self.number_of_digits != 10 or self.recognizer_type != "neural net":
                print(
                    "Using real images; forcing number of digits to 10 and recognizer type to neural net"
                )
            self.number_of_digits = 10
            self.recognizer_type = "neural net"

        if self.recognizer_type == "uniform":
            self.recognizer_type = "random table"
            self.upper_bound_for_log_potential_in_random_table = (
                0  # A value of 0 samples from [0, 0], providing a uniform table.
            )

    # override
    def learning_is_needed(self):
        return self.recognizer_type != "fixed ground truth table"

    # Section: making the data loader

    # override
    def make_data_loader(self):
        self.setup_images()

        if self.custom_digits_and_constraints_values_batches_generator is not None:
            batch_generator = self.generate_custom_examples_batch
        else:
            batch_generator = self.generate_random_examples_batch

        train_data_loader = data_loader_from_batch_generator(
            self.number_of_batches_per_epoch, batch_generator, print=None
        )

        return train_data_loader

    def setup_images(self):
        if self.use_real_images:
            self.setup_real_images()
        else:
            self.setup_fake_images()

    def setup_real_images(self):
        self.images_by_digits_by_phase = read_mnist(self.max_real_mnist_datapoints)
        number_of_training_images = sum(
            [
                len(self.images_by_digits_by_phase["train"][d])
                for d in range(self.number_of_digits)
            ]
        )
        print(f"Loaded {number_of_training_images:,} training images")
        self.next_image_index_by_digit = {d: 0 for d in range(self.number_of_digits)}
        self.show_digit_examples()
        self.get_image_batch_from_digit_batch = self.get_next_real_image_batch_for_digit_batch

    def show_digit_examples(self):
        if self.show_examples:
            images = [
                self.images_by_digits_by_phase["train"][d][i]
                for i in range(5)
                for d in range(self.number_of_digits)
            ]
            labels = [d for i in range(5) for d in range(self.number_of_digits)]
            show_images_and_labels(5, 10, images, labels)

    def setup_fake_images(self):
        self.get_image_batch_from_digit_batch = self.get_next_fake_image_batch_for_digit_batch

    def get_next_fake_image_batch_for_digit_batch(self, digit_batch):
        return digit_batch

    def get_next_real_image_batch_for_digit_batch(self, digit_batch):
        images_list = []
        for d in digit_batch:
            d = d.item()
            image = self.images_by_digits_by_phase["train"][d][self.next_image_index_by_digit[d]]
            if self.use_a_single_image_per_digit:
                pass  # leave the index at the first position forever
            else:
                self.next_image_index_by_digit[d] += 1
                if self.next_image_index_by_digit[d] == len(
                        self.images_by_digits_by_phase["train"][d]
                ):
                    self.next_image_index_by_digit[d] = 0
            images_list.append(image)
        images_batch = torch.stack(images_list).to(digit_batch.device)
        return images_batch

    ### Generation of random examples

    def generate_random_examples_batch(self):
        return self.generate_examples_batch_from_digits_and_constraints_values_generator(
            self.generate_random_digits_and_constraints_values_batches)

    def generate_random_digits_and_constraints_values_batches(self):
        digit_values = self.generate_random_digits_values_batch()
        constraint_values = [
            self.ith_constraint_values(i, digit_values)
            for i in range(self.number_of_constraints)
        ]
        return digit_values , constraint_values

    def generate_random_digits_values_batch(self):
        digit_values = [
            torch.randint(self.number_of_digits, (self.batch_size,)) for i in range(self.chain_length)
        ]
        return digit_values

    def ith_constraint_values(self, i, digit_values):
        digit_arguments = [digit_values[j] for j in self.indices_of_digit_arguments_of_constraint(i)]
        return self.constraint_function(i, *digit_arguments).long()

    ### Custom generation of examples

    def generate_custom_examples_batch(self):
        return self.generate_examples_batch_from_digits_and_constraints_values_generator(
            self.generate_custom_digits_and_constraints_values_batches)

    def generate_custom_digits_and_constraints_values_batches(self):
        return self.custom_digits_and_constraints_values_batches_generator(
            self.number_of_digits, self.chain_length, self.number_of_constraints, self.batch_size
        )

    ### Generation of batches

    def generate_examples_batch_from_digits_and_constraints_values_generator(
            self, digits_and_constraints_values_generator):

        digit_values, constraint_values = digits_and_constraints_values_generator()

        image_values = [
            self.get_image_batch_from_digit_batch(digit_value)
            for digit_value in digit_values
        ]
        observation_dict = {
            self.image_random_variables[i]: image_value
            for i, image_value in enumerate(image_values)
        }
        query_assignment_dict = {
            self.constraint_variables[i]: constraint_values[i]
            for i, constraint_value in enumerate(constraint_values)
        }
        return observation_dict, query_assignment_dict

    # Section: making the model

    # override
    def setup_model(self):
        self.create_random_variables()
        self.recognizer_factors = self.make_recognizer_factors()
        self.constraint_factors = self.make_constraint_factors()
        self.model = [
            # IMPORTANT: this particular factor order is relied upon later in the code
            *self.recognizer_factors,
            *self.constraint_factors,
        ]
        print("\nInitial model:")
        print(join(self.model, "\n"))

    def create_random_variables(self):
        self.image_random_variables = []
        self.digit_variables = []
        self.constraint_variables = []
        for i in range(self.chain_length):
            self.image_random_variables.append(
                TensorVariable(f"image{i}", non_batch_dim=2)
                if self.use_real_images
                else IntegerVariable(f"image{i}", self.number_of_digits)
            )
            self.digit_variables.append(IntegerVariable(f"digit{i}", self.number_of_digits))
            if i != self.number_of_constraints:
                self.constraint_variables.append(IntegerVariable(f"constraint{i}", self.number_of_constraint_values))

    def make_recognizer_factors(self):
        if self.recognizer_type == "neural net":
            if self.use_real_images:
                if self.use_conv_net:
                    print("Using ConvNet")
                    neural_net_maker = lambda: FromLogToProbabilitiesAdapter(ConvNet())
                else:
                    print("Using MLP for MNIST")

                    def neural_net_maker():
                        net = MLP(28 * 28, self.number_of_digits, self.number_of_digits)
                        return net

            else:
                neural_net_maker = lambda: MLP(1, self.number_of_digits, self.number_of_digits)

            def make_inner_function():
                neural_net = neural_net_maker()
                self.uniform_pre_training(neural_net)
                return neural_net

            def make_recognizer_factor(i, neural_net):
                return NeuralFactor(
                    neural_net, input_variables=[self.image_random_variables[i]],
                    output_variable=self.digit_variables[i]
                )

        elif self.recognizer_type == "fixed ground truth table":
            predicate = lambda i, d: d == i

            def make_inner_function():
                return PyTorchTableFactor.from_predicate(
                    [self.image_random_variables[0], self.digit_variables[0]], predicate, log_space=True
                ).table

            def make_recognizer_factor(i, image_and_digit_table):
                return PyTorchTableFactor([self.image_random_variables[i], self.digit_variables[i]],
                                          image_and_digit_table)

        elif self.recognizer_type == "noisy left-shift":
            probability_of_left_shift = 1 - self.left_shift_noise
            number_of_non_left_shift = self.number_of_digits - 1
            probability_of_each_non_left_shift = self.left_shift_noise / number_of_non_left_shift
            left_shift_pairs = {
                (i, (i - 1) % self.number_of_digits) for i in range(self.number_of_digits)
            }

            def potential(i, d):
                return (
                    probability_of_left_shift
                    if (i, d) in left_shift_pairs
                    else probability_of_each_non_left_shift
                )

            def make_inner_function():
                return PyTorchTableFactor.from_function(
                    [self.image_random_variables[0], self.digit_variables[0]], potential, log_space=True
                ).table

            def make_recognizer_factor(i, image_and_digit_table):
                return PyTorchTableFactor([self.image_random_variables[i], self.digit_variables[i]],
                                          image_and_digit_table)

        elif self.recognizer_type == "random table":

            def make_random_parameters():
                return (
                        torch.rand(self.number_of_digits, self.number_of_digits)
                        * self.upper_bound_for_log_potential_in_random_table
                ).requires_grad_(True)

            def make_inner_function():
                return PyTorchLogTable(make_random_parameters())

            def make_recognizer_factor(i, image_and_digit_table):
                return PyTorchTableFactor([self.image_random_variables[i], self.digit_variables[i]],
                                          image_and_digit_table)

        else:
            raise Exception(f"Unknown recognizer type: {self.recognizer_type}")

        inner_functions = []
        for i in range(self.chain_length):
            inner_function = (
                inner_functions[0]
                if i != 0 and self.use_shared_recognizer
                else make_inner_function()
            )
            inner_functions.append(inner_function)

        recognizer_factors = [
            make_recognizer_factor(i, inner_function)
            for i, inner_function in enumerate(inner_functions)
        ]

        return recognizer_factors

    def make_constraint_factors(self):
        constraint_factors = [
            FixedPyTorchTableFactor.from_predicate(
                self.make_constraint_random_variable_parameters(i, self.digit_variables, self.constraint_variables),
                self.make_constraint_predicate(i)
            )
            for i in range(self.number_of_constraints)
        ]
        return constraint_factors

    def make_constraint_random_variable_parameters(self, constraint_index, digit_variables, constraint_variables):
        indices_of_digit_variables = self.indices_of_digit_arguments_of_constraint(constraint_index)
        argument_digit_variables = [digit_variables[j] for j in indices_of_digit_variables]
        argument_constraint_variables = [constraint_variables[constraint_index]]
        return argument_constraint_variables + argument_digit_variables

    def make_constraint_predicate(self, constraint_index):
        return \
            lambda constraint_value, *constraint_digit_arguments: \
                int(self.constraint_function(constraint_index, *constraint_digit_arguments)) == constraint_value

    # Section: Evaluation

    # override
    def after_epoch(self, learner):
        print()
        default_after_epoch(learner)
        if learner.epoch % self.number_of_epochs_between_evaluations == 0:
            self.print_evaluation(learner)

    # override
    def print_evaluation(self, learner=None):
        index_of_first_recognizer_factor_in_model = (
            0
        )  # recognizer factors come first in model, so the first one has index 0
        recognizer_factor = lambda i: self.model[i + index_of_first_recognizer_factor_in_model]
        from_i_to_d = [
            lambda image: recognizer_factor(i)
                .condition({self.image_random_variables[i]: image})
                .normalize()
                .table_factor
            for i in range(self.chain_length)
        ]
        with torch.no_grad():
            predictors = [from_i_to_d[0]] if self.use_shared_recognizer else from_i_to_d
            for predictor in predictors:
                self.print_posterior_of(predictor, learner)

    def print_posterior_of(self, predictor, learner=None):
        space_header = " " * len(self.make_digit_header(0))
        print(space_header + ", ".join([f"   {i}" for i in range(self.number_of_digits)]))
        for digit in range(self.number_of_digits):
            digit_batch = torch.full((self.number_of_digit_instances_in_evaluation,), digit)
            image_batch = self.get_image_batch_from_digit_batch(digit_batch)
            if learner is not None and learner.device is not None:
                image_batch = image_batch.to(learner.device)
            posterior_probability = predictor(image_batch)
            if posterior_probability.batch:
                posterior_probability_tensor = (
                    posterior_probability.table.potentials_tensor().sum(0)
                )
                posterior_probability_tensor /= posterior_probability_tensor.sum()
            else:
                posterior_probability_tensor = (
                    posterior_probability.table.potentials_tensor()
                )
            self.print_posterior_tensor(digit, posterior_probability_tensor)

    def print_posterior_tensor(self, digit, output_probability_tensor):
        digit_header = self.make_digit_header(digit)
        print(digit_header + f"{self.digit_distribution_tensor_str(output_probability_tensor)}")

    def make_digit_header(self, digit):
        image_description = "image" if self.use_real_images else 'fake "image"'
        digit_header = f"Prediction for {image_description} {digit}: "
        return digit_header

    def digit_distribution_tensor_str(self, tensor):
        return join([self.potential_str(potential) for potential in tensor])

    def potential_str(self, potential):
        if potential < 1e-2:
            return " " * (2 + self.evaluation_probability_precision)
        else:
            return f"{potential:0.{self.evaluation_probability_precision}f}"

    def uniform_pre_training(self, recognizer):
        def random_batch_generator():
            if self.use_real_images:
                from_digit_batch_to_image_batch = self.get_next_real_image_batch_for_digit_batch
            else:
                from_digit_batch_to_image_batch = self.get_next_fake_image_batch_for_digit_batch

            def generator():
                digits = torch.randint(self.number_of_digits, (self.batch_size,))
                images = from_digit_batch_to_image_batch(digits)
                return images

            return generator

        print("Uniform pre-training")
        print("Model:")
        print(recognizer)
        model = recognizer
        data_loader = data_loader_from_batch_generator(
            self.number_of_batches_per_epoch, random_batch_generator(), print=None
        )
        UniformTraining(model, data_loader, self.number_of_digits).learn()
        print("Uniform pre-training completed")
        print("Model:")
        print(recognizer)


# Section: execution

def solve_learning_problem_from_parameters(parameters):
    set_seed(parameters.seed)
    solve_learning_problem(
        MNISTChainsProblem(
            number_of_digits=parameters.number_of_digits,
            chain_length=parameters.chain_length,
            number_of_constraints=parameters.number_of_constraints,
            number_of_constraint_values=parameters.number_of_constraint_values,

            indices_of_digit_arguments_of_constraint=parameters.indices_of_digit_arguments_of_constraint,
            constraint_function=parameters.constraint_function,
            custom_digits_and_constraints_values_batches_generator=parameters.custom_digits_and_constraints_values_batches_generator,

            use_real_images=parameters.use_real_images,
            use_conv_net=parameters.use_conv_net,

            show_examples=parameters.show_examples,
            use_a_single_image_per_digit=parameters.use_a_single_image_per_digit,
            batch_size=parameters.batch_size,
            number_of_batches_between_updates=parameters.number_of_batches_between_updates,
            number_of_batches_per_epoch=parameters.number_of_batches_per_epoch,
            number_of_epochs_between_evaluations=parameters.number_of_epochs_between_evaluations,
            max_real_mnist_datapoints=parameters.max_real_mnist_datapoints,
            number_of_digit_instances_in_evaluation=parameters.number_of_digit_instances_in_evaluation,
            seed=parameters.seed,
            evaluation_probability_precision=parameters.evaluation_probability_precision,

            use_shared_recognizer=parameters.use_shared_recognizer,
            recognizer_type=parameters.recognizer_type,
            left_shift_noise=parameters.left_shift_noise,
            upper_bound_for_log_potential_in_random_table=parameters.upper_bound_for_log_potential_in_random_table,
        ),
        try_cuda=parameters.try_cuda,
        lr=parameters.lr,
        loss_decrease_tol=parameters.loss_decrease_tol,
        max_epochs_to_go_before_stopping_due_to_loss_decrease=parameters.max_epochs_to_go_before_stopping_due_to_loss_decrease
    )
